"""Meteorology data models."""

from __future__ import annotations

import hashlib
import json
import logging
import pathlib
import typing
import warnings
from abc import ABC, abstractmethod
from contextlib import ExitStack
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Hashable,
    Iterable,
    Iterator,
    Literal,
    Mapping,
    MutableMapping,
    Sequence,
    TypeVar,
    overload,
)

import numpy as np
import numpy.typing as npt
import xarray as xr
from overrides import overrides

from pycontrails.core import interpolation
from pycontrails.core import vector as vector_module
from pycontrails.core.cache import CacheStore, DiskCacheStore
from pycontrails.core.met_var import AirPressure, Altitude, MetVariable
from pycontrails.physics import units
from pycontrails.utils import temp as temp_module

logger = logging.getLogger(__name__)

# optional imports
if TYPE_CHECKING:
    import open3d as o3d

XArrayType = TypeVar("XArrayType", xr.Dataset, xr.DataArray)
MetDataType = TypeVar("MetDataType", "MetDataset", "MetDataArray")
DatasetType = TypeVar("DatasetType", xr.Dataset, "MetDataset")

COORD_DTYPE = np.float64


class MetBase(ABC, Generic[XArrayType]):
    """Abstract class for building Meteorology Data handling classes.

    All support here should be generic to work on xr.DataArray
    and xr.Dataset.
    """

    #: DataArray or Dataset
    data: XArrayType

    #: Cache datastore to use for :meth:`save` or :meth:`load`
    cachestore: CacheStore | None

    #: Default dimension order for DataArray or Dataset (x, y, z, t)
    dim_order: list[Hashable] = [
        "longitude",
        "latitude",
        "level",
        "time",
    ]

    def __repr__(self) -> str:
        data = getattr(self, "data", None)
        return (
            f"{self.__class__.__name__} with data:\n\n{data.__repr__() if data is not None else ''}"
        )

    def _repr_html_(self) -> str:
        try:
            return f"<b>{type(self).__name__}</b> with data:<br/ ><br/> {self.data._repr_html_()}"
        except AttributeError:
            return f"<b>{type(self).__name__}</b> without data"

    def _validate_dim_contains_coords(self) -> None:
        """Check that data contains four temporal-spatial coordinates.

        Raises
        ------
        ValueError
            If data does not contain all four coordinates (longitude, latitude, level, time).
        """
        for dim in self.dim_order:
            if dim not in self.data.dims:
                if dim == "level":
                    raise ValueError(
                        f"Meteorology data must contain dimension '{dim}'. "
                        "For single level data, set 'level' coordinate to constant -1 "
                        "using `ds = ds.expand_dims({'level': [-1]})`"
                    )
                else:
                    raise ValueError(f"Meteorology data must contain dimension '{dim}'.")

    def _validate_longitude(self) -> None:
        """Check longitude bounds.

        Assumes ``longitude`` dimension is already sorted.

        Raises
        ------
        ValueError
            If longitude values are not contained in the interval [-180, 180].
        """
        longitude = self.variables["longitude"].values
        if longitude.dtype != COORD_DTYPE:
            raise ValueError(
                "Longitude values must be of type float64. "
                "Initiate with 'copy=True' to convert to float64. "
                "Initiate with 'validate=False' to skip validation."
            )

        if self.is_wrapped:
            # Relax verification if the longitude has already been processed and wrapped
            if longitude[-1] > 360.0:
                raise ValueError(
                    "Longitude contains values > 360. Shift to WGS84 with "
                    "'data.assign_coords(longitude=(((data.longitude + 180) % 360) - 180))'"
                )
            if longitude[0] < -360.0:
                raise ValueError(
                    "Longitude contains values < -360. Shift to WGS84 with "
                    "'data.assign_coords(longitude=(((data.longitude + 180) % 360) - 180))'"
                )
            return

        # Strict!
        if longitude[-1] > 180.0:
            raise ValueError(
                "Longitude contains values > 180. Shift to WGS84 with "
                "'data.assign_coords(longitude=(((data.longitude + 180) % 360) - 180))'"
            )
        if longitude[0] < -180.0:
            raise ValueError(
                "Longitude contains values < -180. Shift to WGS84 with "
                "'data.assign_coords(longitude=(((data.longitude + 180) % 360) - 180))'"
            )

    def _validate_latitude(self) -> None:
        """Check latitude bounds.

        Assumes ``latitude`` dimension is already sorted.

        Raises
        ------
        ValueError
            If latitude values are not contained in the interval [-90, 90].
        """
        latitude = self.variables["latitude"].values
        if latitude.dtype != COORD_DTYPE:
            raise ValueError(
                "Latitude values must be of type float64. "
                "Initiate with 'copy=True' to convert to float64. "
                "Initiate with 'validate=False' to skip validation."
            )

        if latitude[0] < -90.0:
            raise ValueError(
                "Latitude contains values < -90 . "
                "Latitude values must be contained in the interval [-90, 90]."
            )
        if latitude[-1] > 90.0:
            raise ValueError(
                "Latitude contains values > 90 . "
                "Latitude values must be contained in the interval [-90, 90]."
            )

    def _validate_sorting(self) -> None:
        """Check that all coordinates are sorted.

        Raises
        ------
        ValueError
            If one of the coordinates is not sorted.
        """
        if not np.all(np.diff(self.variables["time"]) > np.timedelta64(0, "ns")):
            raise ValueError("Coordinate `time` not sorted. Initiate with `copy=True`.")
        for coord in self.dim_order[:3]:  # exclude time, the 4th dimension
            if not np.all(np.diff(self.variables[coord]) > 0.0):
                raise ValueError(f"Coordinate '{coord}' not sorted. Initiate with 'copy=True'.")

    def _validate_transpose(self) -> None:
        """Check that data is transposed according to :attr:`dim_order`."""

        dims_tuple = tuple(self.dim_order)

        def _check_da(da: xr.DataArray, key: str | None = None) -> None:
            if da.dims != dims_tuple:
                if key is not None:
                    msg = (
                        "Data dimension not transposed on variable '{key}'. Initiate with"
                        " `copy=True`."
                    )
                else:
                    msg = "Data dimension not transposed. Initiate with `copy=True`."
                raise ValueError(msg)

        data = self.data
        if isinstance(data, xr.DataArray):
            _check_da(data)
            return

        for key, da in self.data.data_vars.items():
            _check_da(da, key)

    def _validate_dims(self) -> None:
        """Apply all validators."""
        self._validate_dim_contains_coords()

        # Apply this one first: validate_longitude and validate_latitude assume sorted
        self._validate_sorting()
        self._validate_longitude()
        self._validate_latitude()
        self._validate_transpose()

    def _preprocess_dims(self, wrap_longitude: bool) -> None:
        """Confirm DataArray or Dataset include required dimension in a consistent format.

        Expects DataArray or Dataset to contain dimensions ``latitude`, ``longitude``, ``time``,
        and ``level`` (in hPa/mbar).
        Adds additional coordinate variables ``air_pressure`` and ``altitude`` coordinates
        mapped to "level" dimension if "level" > 0.

        Set ``level`` to -1 to signify single level.

        .. versionchanged:: 0.40.0

            All coordinate data (longitude, latitude, level) are promoted to ``float64``.
            Auxiliary coordinates (altitude and air_pressure) are now cast to the same
            dtype as the underlying grid data.


        Parameters
        ----------
        wrap_longitude : bool
            If True, ensure longitude values cover the interval ``[-180, 180]``.

        Raises
        ------
        ValueError
            Raises if required dimension names are not found
        """
        self._validate_dim_contains_coords()

        # Ensure spatial coordinates all have dtype COORD_DTYPE
        for coord in ("longitude", "latitude", "level"):
            arr = self.variables[coord].values
            if arr.dtype != COORD_DTYPE:
                self.data[coord] = arr.astype(COORD_DTYPE)

        # Ensure time is np.datetime64[ns]
        self.data["time"] = self.data["time"].astype("datetime64[ns]", copy=False)

        # sortby to ensure each coordinate has ascending order
        self.data = self.data.sortby(self.dim_order, ascending=True)

        if not self.is_wrapped:
            # Ensure longitude is contained in interval [-180, 180)
            # If longitude has value at 180, we might not want to shift it?
            lon = self.variables["longitude"].values

            # This longitude shifting can give rise to precision errors with float32
            # Only shift if necessary
            if np.any(lon >= 180.0) or np.any(lon < -180.0):
                self.data = shift_longitude(self.data)
            else:
                self.data = self.data.sortby("longitude", ascending=True)

            # wrap longitude, if requested
            if wrap_longitude:
                self.data = _wrap_longitude(self.data)

        self._validate_longitude()
        self._validate_latitude()

        # transpose to have ordering (x, y, z, t, ...)
        dim_order = self.dim_order + [d for d in self.data.dims if d not in self.dim_order]
        self.data = self.data.transpose(*dim_order)

        # single level data
        if self.is_single_level:
            # add level attributes to reflect surface level
            self.data["level"].attrs.update(units="", long_name="Single Level")
            return

        # pressure level data
        level = self.variables["level"].values

        # add pressure level attributes
        self.data["level"].attrs.update(units="hPa", long_name="Pressure", positive="down")

        # add altitude and air_pressure

        # XXX: use the dtype of the data to determine the precision of these coordinates
        # There are two competing conventions here:
        # - coordinate data should be float64
        # - gridded data is typically float32
        # - air_pressure and altitude often play both roles
        # It is more important for air_pressure and altitude to be grid-aligned than to be
        # coordinate-aligned, so we use the dtype of the data to determine the precision of
        # these coordinates
        if isinstance(self.data, xr.Dataset):
            dtype = np.result_type(*self.data.data_vars.values(), np.float32)
        else:
            dtype = self.data.dtype

        level = level.astype(dtype)
        air_pressure = level * 100.0
        altitude = units.pl_to_m(level)
        self.data = self.data.assign_coords({"air_pressure": ("level", air_pressure)})
        self.data = self.data.assign_coords({"altitude": ("level", altitude)})

        # add air_pressure units and long name attributes
        self.data.coords["air_pressure"].attrs.update(
            standard_name=AirPressure.standard_name,
            long_name=AirPressure.long_name,
            units=AirPressure.units,
        )
        # add altitude units and long name attributes
        self.data.coords["altitude"].attrs.update(
            standard_name=Altitude.standard_name,
            long_name=Altitude.long_name,
            units=Altitude.units,
        )

    @property
    def hash(self) -> str:
        """Generate a unique hash for this met instance.

        Note this is not as robust as it could be since `repr`
        cuts off.

        Returns
        -------
        str
            Unique hash for met instance (sha1)
        """
        _hash = repr(self.data)
        return hashlib.sha1(bytes(_hash, "utf-8")).hexdigest()

    @property
    @abstractmethod
    def size(self) -> int:
        """Return the size of (each) array in underlying :attr:`data`.

        Returns
        -------
        int
            Total number of grid points in underlying data
        """

    @property
    @abstractmethod
    def shape(self) -> tuple[int, int, int, int]:
        """Return the shape of the dimensions.

        Returns
        -------
        tuple[int, int, int, int]
            Shape of underlying data
        """

    @property
    def coords(self) -> dict[str, np.ndarray]:
        """Get coordinates of underlying :attr:`data` coordinates.

        Only return non-dimension coordinates.

        See:
        http://xarray.pydata.org/en/stable/user-guide/data-structures.html#coordinates

        Returns
        -------
        dict[str, np.ndarray]
            Dictionary of coordinates
        """
        return {
            "longitude": self.variables["longitude"].values,
            "latitude": self.variables["latitude"].values,
            "level": self.variables["level"].values,
            "time": self.variables["time"].values,
        }

    @property
    def variables(self) -> dict[Hashable, xr.Variable]:
        """Low level access to underlying :attr:`data` variables.

        This method is typically is faster for accessing coordinate variables.

        .. versionadded:: 0.25.2

        Returns
        -------
        dict[Hashable, xr.Variable]
            Dictionary of variables. The type is actually..

                xarray.core.utils.Frozen[Any, xr.Variable]

            In practice, this behaves like a dictionary.

        Examples
        --------
        >>> from pycontrails.datalib.ecmwf import ERA5
        >>> times = (datetime(2022, 3, 1, 12),  datetime(2022, 3, 1, 13))
        >>> variables = "air_temperature", "specific_humidity"
        >>> levels = [200, 300]
        >>> era5 = ERA5(times, variables, levels)
        >>> mds = era5.open_metdataset()
        >>> mds.variables["level"].values  # faster access than mds.data["level"]
        array([200., 300.])

        >>> mda = mds["air_temperature"]
        >>> mda.variables["level"].values  # faster access than mda.data["level"]
        array([200., 300.])
        """
        if isinstance(self.data, xr.Dataset):
            return self.data.variables  # type: ignore[return-value]
        return self.data.coords.variables

    @property
    def is_wrapped(self) -> bool:
        """Check if the longitude dimension covers the closed interval ``[-180, 180]``.

        Assumes the longitude dimension is sorted (this is established by the
        :class:`MetDataset` or :class:`MetDataArray` constructor).

        .. versionchanged 0.26.0::

            The previous implementation checked for the minimum and maximum longitude
            dimension values to be duplicated. The current implementation only checks for
            that the interval ``[-180, 180]`` is covered by the longitude dimension. The
            :func:`pycontrails.physics.geo.advect_longitude` is designed for compatibility
            with this convention.

        Returns
        -------
        bool
            True if longitude coordinates cover ``[-180, 180]``

        See Also
        --------
        :func:`pycontrails.physics.geo.advect_longitude`
        """
        longitude = self.variables["longitude"].values
        return _is_wrapped(longitude)

    @property
    def is_single_level(self) -> bool:
        """Check if instance contains "single level" or "surface level" data.

        This method checks if ``level`` dimension contains a single value equal
        to -1, the pycontrails convention for surface only data.

        Returns
        -------
        bool
            If instance contains single level data.
        """
        level = self.variables["level"].values
        return len(level) == 1 and level[0] == -1

    @abstractmethod
    def broadcast_coords(self, name: str) -> xr.DataArray:
        """Broadcast coordinates along other dimensions.

        Parameters
        ----------
        name : str
            Coordinate/dimension name to broadcast.
            Can be a dimension or non-dimension coordinates.

        Returns
        -------
        xr.DataArray
            DataArray of the coordinate broadcasted along all other dimensions.
            The DataArray will have the same shape as the gridded data.
        """

    def _save(self, dataset: xr.Dataset, **kwargs: Any) -> list[str]:
        """Save dataset to netcdf files named for the met hash and hour.

        Does not yet save in parallel.

        ..versionchanged::0.34.1

            If :attr:`cachestore` is None, this method assigns it
            to new :class:`DiskCacheStore`.

        Parameters
        ----------
        dataset : xr.Dataset
            Dataset to save
        **kwargs
            Keyword args passed directly on to :func:`xarray.save_mfdataset`

        Returns
        -------
        list[str]
            List of filenames saved
        """
        self.cachestore = self.cachestore or DiskCacheStore()

        # group by hour and save one dataset for each hour to temp file
        times, datasets = zip(*dataset.groupby("time", squeeze=False))

        # Open ExitStack to control temp_file context manager
        with ExitStack() as stack:
            xarray_temp_filenames = [stack.enter_context(temp_module.temp_file()) for _ in times]
            xr.save_mfdataset(datasets, xarray_temp_filenames, **kwargs)

            # set filenames by hash
            filenames = [f"{self.hash}-{t_idx}.nc" for t_idx, _ in enumerate(times)]

            # put each hourly file into cache
            self.cachestore.put_multiple(xarray_temp_filenames, filenames)

        return filenames

    def __len__(self) -> int:
        return self.data.__len__()

    @property
    def attrs(self) -> dict[Hashable, Any]:
        """Pass through to `self.data.attrs`."""
        return self.data.attrs

    @abstractmethod
    def downselect(self, bbox: list[float]) -> MetBase:
        """Downselect met data within spatial bounding box.

        Parameters
        ----------
        bbox : list[float]
            List of coordinates defining a spatial bounding box in WGS84 coordinates.
            For 2D queries, list is [west, south, east, north].
            For 3D queries, list is [west, south, min-level, east, north, max-level]
            with level defined in [:math:`hPa`].


        Returns
        -------
        MetBase
            Return downselected data
        """

    @property
    def is_zarr(self) -> bool:
        """Check if underlying :attr:`data` is sourced from a Zarr group.

        Implementation is very brittle, and may break as external libraries change.

        Some ``dask`` intermediate artifact is cached when this is called. Typically,
        subsequent calls to this method are much faster than the initial call.

        .. versionadded:: 0.26.0

        Returns
        -------
        bool
            If ``data`` is based on a Zarr group.
        """
        return _is_zarr(self.data)


class MetDataset(MetBase):
    """Meteorological dataset with multiple variables.

    Composition around xr.Dataset to enforce certain
    variables and dimensions for internal usage

    Parameters
    ----------
    data : xr.Dataset
        :class:`xarray.Dataset` containing meteorological variables and coordinates
    cachestore : :class:`CacheStore`, optional
        Cache datastore for staging intermediates with :meth:`save`.
        Defaults to None.
    wrap_longitude : bool, optional
        Wrap data along the longitude dimension. If True, duplicate and shift longitude
        values (ie, -180 -> 180) to ensure that the longitude dimension covers the entire
        interval ``[-180, 180]``. Defaults to False.
    copy : bool, optional
        Copy data on construction. Defaults to True.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> import xarray as xr
    >>> from pycontrails.datalib.ecmwf import ERA5

    >>> time = ("2022-03-01T00", "2022-03-01T02")
    >>> variables = ["air_temperature", "specific_humidity"]
    >>> pressure_levels = [200, 250, 300]
    >>> era5 = ERA5(time, variables, pressure_levels)

    >>> # Open directly as `MetDataset`
    >>> met = era5.open_metdataset()
    >>> # Use `data` attribute to access `xarray` object
    >>> assert isinstance(met.data, xr.Dataset)

    >>> # Alternatively, open with `xarray` and cast to `MetDataset`
    >>> ds = xr.open_mfdataset(era5._cachepaths)
    >>> met = MetDataset(ds)

    >>> # Access sub-`DataArrays`
    >>> mda = met["t"]  # `MetDataArray` instance, needed for interpolation operations
    >>> da = mda.data  # Underlying `xarray` object

    >>> # Check out a few values
    >>> da[5:10, 5:10, 1, 1].values
    array([[224.0896 , 224.41374, 224.75946, 225.16237, 225.60507],
           [224.09457, 224.42038, 224.76526, 225.16817, 225.61089],
           [224.10037, 224.42618, 224.77106, 225.17314, 225.61586],
           [224.10617, 224.43282, 224.7777 , 225.17812, 225.62166],
           [224.11115, 224.44028, 224.7835 , 225.18393, 225.62663]],
          dtype=float32)

    >>> # Mean temperature over entire array
    >>> da.mean().load().item()
    223.5083
    """

    data: xr.Dataset

    def __init__(
        self,
        data: xr.Dataset,
        cachestore: CacheStore | None = None,
        wrap_longitude: bool = False,
        copy: bool = True,
    ):
        # init cache
        self.cachestore = cachestore

        # if input is already a Dataset, copy into data
        if not isinstance(data, xr.Dataset):
            raise ValueError("Input `data` must be an xarray Dataset")

        # copy Dataset into data
        if copy:
            self.data = data.copy()
            self._preprocess_dims(wrap_longitude)

        else:
            if wrap_longitude:
                raise ValueError("Set 'copy=True' when using 'wrap_longitude=True'.")
            self.data = data
            self._validate_dims()

    def __getitem__(self, key: Hashable) -> MetDataArray:
        """Return DataArray of variable `key` cast to a `MetDataArray` object.

        Parameters
        ----------
        key : Hashable
            Variable name

        Returns
        -------
        MetDataArray
            MetDataArray instance associated to :attr:`data` variable `key`

        Raises
        ------
        KeyError
            If `key` not found in :attr:`data`
        """
        try:
            da = self.data[key]
        except KeyError as e:
            raise KeyError(
                f"Variable {key} not found. Available variables: {', '.join(self.data.data_vars)}. "
                "To get items (e.g. `time` or `level`) from underlying `xr.Dataset` object, "
                "use the `data` attribute."
            ) from e
        return MetDataArray(da, copy=False, validate=False)

    def get(self, key: str, default_value: Any = None) -> Any:
        """Shortcut to :meth:`data.get(k, v)` method.

        Parameters
        ----------
        key : str
            Key to get from :attr:`data`
        default_value : Any, optional
            Return `default_value` if `key` not in :attr:`data`, by default `None`

        Returns
        -------
        Any
            Values returned from  :attr:`data.get(key, default_value)`
        """
        return self.data.get(key, default_value)

    def __setitem__(
        self,
        key: Hashable | list[Hashable] | Mapping,
        value: MetDataArray | xr.DataArray | np.ndarray | xr.Variable,
    ) -> None:
        """Shortcut to set data variable on :attr:`data`.

        Warns if ``key`` is already present in dataset.

        Parameters
        ----------
        key : Hashable | list[Hashable] | Mapping
            Variable name
        value : MetDataArray | ArrayLike | xr.Variable
            Value to set to variable names

        See Also
        --------
        - :class:`xarray.Dataset.__setitem__`
        """

        # pull data of MetDataArray value
        if isinstance(value, MetDataArray):
            value = value.data

        # warn if key is already in Dataset
        override_keys: list[Hashable] = []

        if isinstance(key, Hashable):
            if key in self:
                override_keys = [key]

        # xarray.core.utils.is_dict_like
        # https://github.com/pydata/xarray/blob/4cae8d0ec04195291b2315b1f21d846c2bad61ff/xarray/core/utils.py#L244
        elif xr.core.utils.is_dict_like(key) or isinstance(key, list):
            override_keys = [k for k in key if k in self.data]

        if override_keys:
            warnings.warn(
                f"Overwriting data in keys `{override_keys}`. "
                "Use `.update(...)` to suppress warning."
            )

        self.data.__setitem__(key, value)

    def update(self, other: MutableMapping | None = None, **kwargs: Any) -> None:
        """Shortcut to :meth:`data.update`.

        See :meth:`xarray.Dataset.update` for reference.

        Parameters
        ----------
        other : MutableMapping
            Variables with which to update this dataset
        **kwargs : Any
            Variables defined by keyword arguments. If a variable exists both in
            ``other`` and as a keyword argument, the keyword argument takes
            precedence.

        See Also
        --------
        - :meth:`xarray.Dataset.update`
        """
        other = other or {}
        other.update(kwargs)

        # pull data of MetDataArray value
        for k, v in other.items():
            if isinstance(v, MetDataArray):
                other[k] = v.data

        self.data.update(other)

    def __iter__(self) -> Iterator[str]:
        """Allow for the use as "key" in self.met, where "key" is a data variable."""
        # From the typing perspective, `iter(self.data)`` returns Hashables (not
        # necessarily strs). In everything we do, we use str variables.
        # If we decide to extend this to support Hashable, we'll also want to
        # change VectorDataset -- the underlying :attr:`data` should then be
        # changed to `dict[Hashable, np.ndarray]`.
        for key in self.data:
            yield str(key)

    def __contains__(self, key: Hashable) -> bool:
        """Check if key `key` is in :attr:`data`.

        Parameters
        ----------
        key : Hashable
            Key to check

        Returns
        -------
        bool
            True if `key` is in :attr:`data`, False otherwise
        """
        return key in self.data

    @property
    @overrides
    def shape(self) -> tuple[int, int, int, int]:
        sizes = self.data.sizes
        return sizes["longitude"], sizes["latitude"], sizes["level"], sizes["time"]

    @property
    @overrides
    def size(self) -> int:
        return np.product(self.shape).item()

    def copy(self) -> MetDataset:
        """Create a copy of the current class.

        Returns
        -------
        MetDataset
            MetDataset copy
        """
        return MetDataset(
            self.data,
            cachestore=self.cachestore,
            copy=True,  # True by default, but being extra explicit
        )

    def ensure_vars(
        self,
        vars: MetVariable | str | Sequence[MetVariable | str | Sequence[MetVariable]],
        raise_error: bool = True,
    ) -> list[str]:
        """Ensure variables exist in xr.Dataset.

        Parameters
        ----------
        vars : MetVariable | str | Sequence[MetVariable | str | list[MetVariable]]
            List of MetVariable (or string key), or individual MetVariable (or string key).
            If ``vars`` contains an element with a list[MetVariable], then
            only one variable in the list must be present in dataset.
        raise_error : bool, optional
            Raise KeyError if data does not contain variables.
            Defaults to True.

        Returns
        -------
        list[str]
            List of met keys verified in MetDataset.
            Returns an empty list if any MetVariable is missing.

        Raises
        ------
        KeyError
            Raises when dataset does not contain variable in ``vars``
        """
        if isinstance(vars, MetVariable):
            vars = [vars]

        met_keys: list[str] = []
        for variable in vars:
            met_key: str | None = None

            # input is a MetVariable or str
            if isinstance(variable, MetVariable):
                if (key := variable.standard_name) in self:
                    met_key = key
            elif isinstance(variable, str):
                if variable in self:
                    met_key = variable

            # otherwise, assume input is an sequence
            # Sequence[MetVariable] means that any variable in list will work
            else:
                for v in variable:
                    if (key := v.standard_name) in self:
                        met_key = key
                        break

            if met_key is None:
                if not raise_error:
                    return []

                if isinstance(variable, MetVariable):
                    raise KeyError(f"Dataset does not contain variable `{variable.standard_name}`")
                if isinstance(variable, list):
                    raise KeyError(
                        "Dataset does not contain one of variables "
                        f"`{[v.standard_name for v in variable]}`"
                    )
                raise KeyError(f"Dataset does not contain variable `{variable}`")

            met_keys.append(met_key)

        return met_keys

    def save(self, **kwargs: Any) -> list[str]:
        """Save intermediate to :attr:`cachestore` as netcdf.

        Load and restore using :meth:`load`.

        Parameters
        ----------
        **kwargs : Any
            Keyword arguments passed directly to :meth:`xarray.Dataset.to_netcdf`

        Returns
        -------
        list[str]
            Returns filenames saved
        """
        return self._save(self.data, **kwargs)

    @classmethod
    def load(
        cls,
        hash: str,
        cachestore: CacheStore | None = None,
        chunks: dict[str, int] | None = None,
    ) -> MetDataset:
        """Load saved intermediate from :attr:`cachestore`.

        Parameters
        ----------
        hash : str
            Saved hash to load.
        cachestore : :class:`CacheStore`, optional
            Cache datastore to use for sourcing files.
            Defaults to DiskCacheStore.
        chunks : dict[str: int], optional
            Chunks kwarg passed to :func:`xarray.open_mfdataset()` when opening files.

        Returns
        -------
        MetDataset
            New MetDataArray with loaded data.
        """
        cachestore = cachestore or DiskCacheStore()
        chunks = chunks or {}
        data = _load(hash, cachestore, chunks)
        return cls(data)

    def wrap_longitude(self) -> MetDataset:
        """Wrap longitude coordinates.

        Returns
        -------
        MetDataset
            Copy of MetDataset with wrapped longitude values.
            Returns copy of current MetDataset when longitude values are already wrapped
        """
        return MetDataset(
            _wrap_longitude(self.data),
            cachestore=self.cachestore,
        )

    @overrides
    def broadcast_coords(self, name: str) -> xr.DataArray:
        da = xr.ones_like(self.data[list(self.data.keys())[0]]) * self.data[name]
        da.name = name

        return da

    @overrides
    def downselect(self, bbox: list[float]) -> MetDataset:
        data = downselect(self.data, bbox)
        return MetDataset(data, cachestore=self.cachestore, copy=False)

    def to_vector(self, transfer_attrs: bool = True) -> vector_module.GeoVectorDataset:
        """Convert a :class:`MetDataset` to a :class:`GeoVectorDataset` by raveling data.

        If :attr:`data` is lazy, it will be loaded.

        Parameters
        ----------
        transfer_attrs : bool, optional
            Transfer attributes from :attr:`data` to output :class:`GeoVectorDataset`.
            By default, True, meaning that attributes are transferred.

        Returns
        -------
        GeoVectorDataset
            Converted :class:`GeoVectorDataset`. The variables on the returned instance
            include all of those on the input instance, plus the four core spatial temporal
            variables.

        Examples
        --------
        >>> from pycontrails.datalib.ecmwf import ERA5
        >>> times = "2022-03-01",  "2022-03-01T01"
        >>> variables = ["air_temperature", "specific_humidity"]
        >>> levels = [250, 200]
        >>> era5 = ERA5(time=times, variables=variables, pressure_levels=levels)
        >>> met = era5.open_metdataset()
        >>> met.to_vector(transfer_attrs=False)
        GeoVectorDataset [6 keys x 4152960 length, 1 attributes]
            Keys: longitude, latitude, level, time, air_temperature, ..., specific_humidity
            Attributes:
            time                [2022-03-01 00:00:00, 2022-03-01 01:00:00]
            longitude           [-180.0, 179.75]
            latitude            [-90.0, 90.0]
            altitude            [10362.848672411146, 11783.938524404566]
            crs                 EPSG:4326

        """
        coords_keys = [str(key) for key in self.data.dims]  # str not in Hashable
        coords_vals = [self.variables[key].values for key in coords_keys]
        coords_meshes = np.meshgrid(*coords_vals, indexing="ij")
        raveled_coords = [mesh.ravel() for mesh in coords_meshes]
        data = dict(zip(coords_keys, raveled_coords))

        vector = vector_module.GeoVectorDataset(data, copy=False)
        for key in self:
            # The call to .values here will load the data if it is lazy
            vector[key] = self[key].data.values.ravel()

        if transfer_attrs:
            # vector.attrs expects keys to be strings .... we'll get an error
            # if we cannot cast here
            vector.attrs.update({str(k): v for k, v in self.attrs.items()})
        return vector

    @classmethod
    def from_coords(
        cls,
        longitude: npt.ArrayLike | float,
        latitude: npt.ArrayLike | float,
        level: npt.ArrayLike | float,
        time: npt.ArrayLike | np.datetime64,
    ) -> MetDataset:
        """Create a :class:`MetDataset` containing a coordinate skeleton from coordinate arrays.

        Parameters
        ----------
        longitude, latitude : npt.ArrayLike | float
            Horizontal coordinates, in [:math:`degrees`]
        level : npt.ArrayLike | float
            Vertical coordinate, in [:math:`hPa`]
        time: npt.ArrayLike | np.datetime64,
            Temporal coordinates, in [:math:`UTC`]. Will be sorted.

        Returns
        -------
        MetDataset
            MetDataset with no variables.

        Examples
        --------
        >>> # Create skeleton MetDataset
        >>> longitude = np.arange(0, 10, 0.5)
        >>> latitude = np.arange(0, 10, 0.5)
        >>> level = [250, 300]
        >>> time = np.datetime64("2019-01-01")
        >>> met = MetDataset.from_coords(longitude, latitude, level, time)
        >>> met
        MetDataset with data:
        <xarray.Dataset>
        Dimensions:       (longitude: 20, latitude: 20, level: 2, time: 1)
        Coordinates:
          * longitude     (longitude) float64 0.0 0.5 1.0 1.5 2.0 ... 8.0 8.5 9.0 9.5
          * latitude      (latitude) float64 0.0 0.5 1.0 1.5 2.0 ... 7.5 8.0 8.5 9.0 9.5
          * level         (level) float64 250.0 300.0
          * time          (time) datetime64[ns] 2019-01-01
            air_pressure  (level) float32 2.5e+04 3e+04
            altitude      (level) float32 1.036e+04 9.164e+03
        Data variables:
            *empty*

        >>> met.shape
        (20, 20, 2, 1)

        >>> met.size
        800

        >>> # Fill it up with some constant data
        >>> met["temperature"] = xr.DataArray(np.full(met.shape, 234.5), coords=met.coords)
        >>> met["humidity"] = xr.DataArray(np.full(met.shape, 0.5), coords=met.coords)
        >>> met
        MetDataset with data:
        <xarray.Dataset>
        Dimensions:       (longitude: 20, latitude: 20, level: 2, time: 1)
        Coordinates:
          * longitude     (longitude) float64 0.0 0.5 1.0 1.5 2.0 ... 8.0 8.5 9.0 9.5
          * latitude      (latitude) float64 0.0 0.5 1.0 1.5 2.0 ... 7.5 8.0 8.5 9.0 9.5
          * level         (level) float64 250.0 300.0
          * time          (time) datetime64[ns] 2019-01-01
            air_pressure  (level) float32 2.5e+04 3e+04
            altitude      (level) float32 1.036e+04 9.164e+03
        Data variables:
            temperature   (longitude, latitude, level, time) float64 234.5 ... 234.5
            humidity      (longitude, latitude, level, time) float64 0.5 0.5 ... 0.5 0.5

        >>> # Convert to a GeoVectorDataset
        >>> vector = met.to_vector()
        >>> vector.dataframe.head()
        longitude  latitude  level       time  temperature  humidity
        0        0.0       0.0  250.0 2019-01-01        234.5       0.5
        1        0.0       0.0  300.0 2019-01-01        234.5       0.5
        2        0.0       0.5  250.0 2019-01-01        234.5       0.5
        3        0.0       0.5  300.0 2019-01-01        234.5       0.5
        4        0.0       1.0  250.0 2019-01-01        234.5       0.5
        """
        input_data = {
            "longitude": longitude,
            "latitude": latitude,
            "level": level,
            "time": time,
        }

        # clean up input into coords
        coords: dict[str, np.ndarray] = {}
        for key, val in input_data.items():
            val = np.asarray(val)
            if val.ndim == 0:
                val = val.reshape(1)
            elif val.ndim > 1:
                raise ValueError(f"{key} has too many dimensions")

            val = np.sort(val)
            if val.size == 0:
                raise ValueError(f"Coordinate {key} must be nonempty.")

            # Check dtypes
            if key == "time":
                val = val.astype("datetime64[ns]", copy=False)
            else:
                val = val.astype(COORD_DTYPE, copy=False)

            coords[key] = val

        return cls(xr.Dataset({}, coords=coords))

    @classmethod
    def from_zarr(
        cls,
        store: Any,
        cache_size: float | int | None = None,
        **kwargs: Any,
    ) -> MetDataset:
        """Create a :class:`MetDataset` from a path to a Zarr store.

        Parameters
        ----------
        store : Any
            Path to Zarr store. Passed into :func:`xarray.open_zarr`.
        cache_size: float | int | None, optional
            If not None, wrap the Zarr store with a :class:`zarr.LRUStoreCache`.
            This parameter is the number of bytes of memory for the memory cache.
            For example, use ``cache_size=4e9`` for a 4 gigabyte memory cache store.
        **kwargs : Any
            Other keyword only arguments passed into :func:`xarray.open_zarr`.

        Returns
        -------
        MetDataset
            MetDataset with data from Zarr store.

        Notes
        -----
        The :class:`zarr.LRUStoreCache` approach is not fully optimized because
        it holds encoded chunks (raw ``bytes`` downloaded from a cloud Zarr store)
        as opposed decoded :class:`np.ndarray`s. See

        - https://github.com/zarr-developers/zarr-python/pull/306
        - https://github.com/zarr-developers/zarr-python/issues/278

        After some prototyping and profiling, it seems best to hold decoded
        :class:`np.ndarray` data directly through ``dask``. See :meth:`values`
        or :meth:`xr.DataArray.load`.
        """
        if cache_size is not None:
            if cache_size < 1e9:
                warnings.warn(
                    "The unit for the 'cache_size' parameter is number of bytes. "
                    "For performance, this should be at least one gigabyte. Set to "
                    "at least 1e9 to avoid this warning."
                )

            try:
                import fsspec
                import zarr
            except ModuleNotFoundError:
                raise ModuleNotFoundError(
                    "The zarr and fsspec packages are required to use cache_size. "
                    "Install them with 'pip install pycontrails[zarr]'."
                )

            mapper = fsspec.get_mapper(store)
            store = zarr.LRUStoreCache(mapper, max_size=cache_size)

        kwargs.setdefault("storage_options", {"read_only": True})
        ds = xr.open_zarr(store, **kwargs)
        return cls(ds)


class MetDataArray(MetBase):
    """Meteorological DataArray of single variable.

    Wrapper around xr.DataArray to enforce certain
    variables and dimensions for internal usage.

    Parameters
    ----------
    data : ArrayLike
        xr.DataArray or other array-like data source.
        When array-like input is provided, input ``**kwargs`` passed directly to
        xr.DataArray constructor.
    cachestore : :class:`CacheStore`, optional
        Cache datastore for staging intermediates with :meth:`save`.
        Defaults to DiskCacheStore.
    wrap_longitude : bool, optional
        Wrap data along the longitude dimension. If True, duplicate and shift longitude
        values (ie, -180 -> 180) to ensure that the longitude dimension covers the entire
        interval ``[-180, 180]``. Defaults to False.
    copy : bool, optional
        Copy `data` parameter on construction, by default `True`. If `data` is lazy-loaded
        via `dask`, this parameter has no effect. If `data` is already loaded into memory,
        a copy of the data (rather than a view) may be created if `True`.
    validate : bool, optional
        Confirm that the parameter `data` has correct specification. This automatically handled
        in the case that `copy=True`. Validation only introduces a very small overhead.
        This parameter should only be set to `False` if working with data derived from an
        existing MetDataset or :class`MetDataArray`. By default `True`.
    **kwargs
        If `data` input is not a xr.DataArray, `data` will be passed
        passed directly to xr.DataArray constructor with these keyword arguments.

    Examples
    --------
    >>> import numpy as np
    >>> import xarray as xr
    >>> rng = np.random.default_rng(seed=456)

    >>> # Cook up random xarray object
    >>> coords = {
    ...     "longitude": np.arange(-20, 20),
    ...     "latitude": np.arange(-30, 30),
    ...     "level": [220, 240, 260, 280],
    ...     "time": [np.datetime64("2021-08-01T12", "ns"), np.datetime64("2021-08-01T16", "ns")]
    ...     }
    >>> da = xr.DataArray(rng.random((40, 60, 4, 2)), dims=coords.keys(), coords=coords)

    >>> # Cast to `MetDataArray` in order to interpolate
    >>> from pycontrails import MetDataArray
    >>> mda = MetDataArray(da)
    >>> mda.interpolate(-11.4, 5.7, 234, np.datetime64("2021-08-01T13"))
    array([0.52358215])

    >>> mda.interpolate(-11.4, 5.7, 234, np.datetime64("2021-08-01T13"), method='nearest')
    array([0.4188465])

    >>> da.sel(longitude=-11, latitude=6, level=240, time=np.datetime64("2021-08-01T12")).item()
    0.41884649899766946
    """

    data: xr.DataArray

    def __init__(
        self,
        data: xr.DataArray,
        cachestore: CacheStore | None = None,
        wrap_longitude: bool = False,
        copy: bool = True,
        validate: bool = True,
        **kwargs: Any,
    ) -> None:
        # init cache
        self.cachestore = cachestore

        # try to create DataArray out of input data and **kwargs
        if not isinstance(data, xr.DataArray):
            data = xr.DataArray(data, **kwargs)

        if copy:
            self.data = data.copy()
            self._preprocess_dims(wrap_longitude)
        else:
            if wrap_longitude:
                raise ValueError("Set 'copy=True' when using 'wrap_longitude=True'.")
            self.data = data
            if validate:
                self._validate_dims()

        # if name is specified in kwargs, overwrite any other name in DataArray
        if "name" in kwargs:
            self.data.name = kwargs["name"]

        # if at this point now "name" exists on DataArray, set default
        if self.data.name is None:
            self.data.name = "met"

    @property
    def values(self) -> np.ndarray:
        """Return underlying numpy array.

        This methods loads :attr:`data` if it is not already in memory.

        Returns
        -------
        np.ndarray
            Underlying numpy array

        See Also
        --------
        - :meth:`xr.Dataset.load`
        - :meth:`xr.DataArray.load`
        """
        if not self.in_memory:
            self._check_memory("Extracting numpy array from")
            self.data = self.data.load()

        return self.data.values

    @property
    def name(self) -> Hashable:
        """Return the DataArray name.

        Returns
        -------
        Hashable
            DataArray name
        """
        return self.data.name

    @property
    def binary(self) -> bool:
        """Determine if all data is a binary value (0, 1).

        Returns
        -------
        bool
            True if all data values are binary value (0, 1)
        """
        return np.array_equal(self.data, self.data.astype(bool))

    @property
    @overrides
    def size(self) -> int:
        return self.data.size

    @property
    @overrides
    def shape(self) -> tuple[int, int, int, int]:
        # https://github.com/python/mypy/issues/1178
        return typing.cast(typing.Tuple[int, int, int, int], self.data.shape)

    def copy(self) -> MetDataArray:
        """Create a copy of the current class.

        Returns
        -------
        MetDataArray
            MetDataArray copy
        """
        return MetDataArray(self.data, cachestore=self.cachestore, copy=True)

    def wrap_longitude(self) -> MetDataArray:
        """Wrap longitude coordinates.

        Returns
        -------
        MetDataArray
            Copy of MetDataArray with wrapped longitude values.
            Returns copy of current MetDataArray when longitude values are already wrapped
        """
        return MetDataArray(_wrap_longitude(self.data), cachestore=self.cachestore)

    @property
    def in_memory(self) -> bool:
        """Check if underlying :attr:`data` is loaded into memory.

        This method uses protected attributes of underlying `xarray` objects, and may be subject
        to deprecation.

        .. versionchanged:: 0.26.0

            Rename from ``is_loaded`` to ``in_memory``.

        Returns
        -------
        bool
            If underlying data exists as an `np.ndarray` in memory.
        """
        return self.data._in_memory

    @overload
    def interpolate(
        self,
        longitude: float | npt.NDArray[np.float_],
        latitude: float | npt.NDArray[np.float_],
        level: float | npt.NDArray[np.float_],
        time: np.datetime64 | npt.NDArray[np.datetime64],
        *,
        method: str = ...,
        bounds_error: bool = ...,
        fill_value: float | np.float64 | None = ...,
        localize: bool = ...,
        indices: interpolation.RGIArtifacts | None = None,
        return_indices: Literal[False] = ...,
    ) -> npt.NDArray[np.float_]:
        ...

    @overload
    def interpolate(
        self,
        longitude: float | npt.NDArray[np.float_],
        latitude: float | npt.NDArray[np.float_],
        level: float | npt.NDArray[np.float_],
        time: np.datetime64 | npt.NDArray[np.datetime64],
        *,
        method: str = ...,
        bounds_error: bool = ...,
        fill_value: float | np.float64 | None = ...,
        localize: bool = ...,
        indices: interpolation.RGIArtifacts | None = None,
        return_indices: Literal[True],
    ) -> tuple[npt.NDArray[np.float_], interpolation.RGIArtifacts]:
        ...

    def interpolate(
        self,
        longitude: float | npt.NDArray[np.float_],
        latitude: float | npt.NDArray[np.float_],
        level: float | npt.NDArray[np.float_],
        time: np.datetime64 | npt.NDArray[np.datetime64],
        *,
        method: str = "linear",
        bounds_error: bool = False,
        fill_value: float | np.float64 | None = np.nan,
        localize: bool = False,
        indices: interpolation.RGIArtifacts | None = None,
        return_indices: bool = False,
    ) -> npt.NDArray[np.float_] | tuple[npt.NDArray[np.float_], interpolation.RGIArtifacts]:
        """Interpolate values over underlying DataArray.

        Zero dimensional coordinates are reshaped to 1D arrays.

        Method automatically loads underlying :attr:`data` into memory.

        If ``method == "nearest"``, the out array will have the same ``dtype`` as
        the underlying :attr:`data`.

        If ``method == "linear"``, the out array will be promoted to the most
        precise ``dtype`` of:

        - underlying :attr:`data`
        - :attr:`data.longitude`
        - :attr:`data.latitude`
        - :attr:`data.level`
        - ``longitude``
        - ``latitude``

        .. versionadded:: 0.24

            This method can now handle singleton dimensions with ``method == "linear"``.
            Previously these degenerate dimensions caused nan values to be returned.

        Parameters
        ----------
        longitude : float | npt.NDArray[np.float_]
            Longitude values to interpolate. Assumed to be 0 or 1 dimensional.
        latitude : float | npt.NDArray[np.float_]
            Latitude values to interpolate. Assumed to be 0 or 1 dimensional.
        level : float | npt.NDArray[np.float_]
            Level values to interpolate. Assumed to be 0 or 1 dimensional.
        time : np.datetime64 | npt.NDArray[np.datetime64]
            Time values to interpolate. Assumed to be 0 or 1 dimensional.
        method: str, optional
            Additional keyword arguments to pass to
            :class:`scipy.interpolate.RegularGridInterpolator`.
            Defaults to "linear".
        bounds_error: bool, optional
            Additional keyword arguments to pass to
            :class:`scipy.interpolate.RegularGridInterpolator`.
            Defaults to ``False``.
        fill_value: float | np.float64, optional
            Additional keyword arguments to pass to
            :class:`scipy.interpolate.RegularGridInterpolator`.
            Set to None to extrapolate outside the boundary when ``method`` is ``nearest``.
            Defaults to ``np.nan``.
        localize: bool, optional
            Experimental. If True, downselect gridded data to smallest bounding box containing
            all points.  By default False.
        indices: tuple | None, optional
            Experimental. See :func:`interpolation.interp`. None by default.
        return_indices: bool, optional
            Experimental. See :func:`interpolation.interp`. False by default.

        Returns
        -------
        np.ndarray
            Interpolated values

        See Also
        --------
        :meth:`GeoVectorDataset.intersect_met`

        Examples
        --------
        >>> from datetime import datetime
        >>> import numpy as np
        >>> import pandas as pd
        >>> from pycontrails.datalib.ecmwf import ERA5

        >>> times = (datetime(2022, 3, 1, 12),  datetime(2022, 3, 1, 15))
        >>> variables = "air_temperature"
        >>> levels = [200, 250, 300]
        >>> era5 = ERA5(times, variables, levels)
        >>> met = era5.open_metdataset()
        >>> mda = met["air_temperature"]

        >>> # Interpolation at a grid point agrees with value
        >>> mda.interpolate(1, 2, 300, np.datetime64('2022-03-01T14:00'))
        array([241.91974], dtype=float32)

        >>> da = mda.data
        >>> da.sel(longitude=1, latitude=2, level=300, time=np.datetime64('2022-03-01T14')).item()
        241.91974

        >>> # Interpolation off grid
        >>> mda.interpolate(1.1, 2.1, 290, np.datetime64('2022-03-01 13:10'))
        array([239.83794], dtype=float32)

        >>> # Interpolate along path
        >>> longitude = np.linspace(1, 2, 10)
        >>> latitude = np.linspace(2, 3, 10)
        >>> level = np.linspace(200, 300, 10)
        >>> time = pd.date_range("2022-03-01T14", periods=10, freq="5T")
        >>> mda.interpolate(longitude, latitude, level, time)
        array([220.44348, 223.089  , 225.7434 , 228.41643, 231.10858, 233.54858,
               235.71506, 237.86479, 239.99275, 242.10793], dtype=float32)
        """
        # Load if necessary
        if not self.in_memory:
            self._check_memory("Interpolation over")
            self.data = self.data.load()

        # Convert all inputs to 1d arrays
        # Not validating against ndim >= 2
        longitude, latitude, level, time = np.atleast_1d(longitude, latitude, level, time)

        # Pass off to the interp function, which does all the heavy lifting
        return interpolation.interp(
            longitude=longitude,
            latitude=latitude,
            level=level,
            time=time,
            da=self.data,
            method=method,
            bounds_error=bounds_error,
            fill_value=fill_value,
            localize=localize,
            indices=indices,
            return_indices=return_indices,
        )

    def _check_memory(self, msg_start: str) -> None:
        n_bytes = self.data.nbytes
        n_gb = n_bytes // int(1e9)
        msg = (
            f"{msg_start} MetDataArray {self.name} requires loading "
            f"at least {n_gb} GB of data into memory. Downselect data if possible. "
            "If working with a GeoVectorDataset instance, this can be achieved "
            "with the method 'downselect_met'."
        )

        if n_gb > 32:  # Prevent something stupid
            raise RuntimeError(msg)
        if n_gb > 4:
            warnings.warn(msg)

        mb = round(n_bytes / int(1e6), 2)
        logger.debug("Loading %s into memory consumes %s MB.", self.name, mb)

    def save(self, **kwargs: Any) -> list[str]:
        """Save intermediate to :attr:`cachestore` as netcdf.

        Load and restore using :meth:`load`.

        Parameters
        ----------
        **kwargs : Any
            Keyword arguments passed directly to :func:`xarray.save_mfdataset`

        Returns
        -------
        list[str]
            Returns filenames of saved files
        """
        dataset = self.data.to_dataset()
        return self._save(dataset, **kwargs)

    @classmethod
    def load(
        cls,
        hash: str,
        cachestore: CacheStore | None = None,
        chunks: dict[str, int] | None = None,
    ) -> MetDataArray:
        """Load saved intermediate from :attr:`cachestore`.

        Parameters
        ----------
        hash : str
            Saved hash to load.
        cachestore : CacheStore, optional
            Cache datastore to use for sourcing files.
            Defaults to DiskCacheStore.
        chunks : dict[str, int], optional
            Chunks kwarg passed to :func:`xarray.open_mfdataset()` when opening files.

        Returns
        -------
        MetDataArray
            New MetDataArray with loaded data.
        """
        cachestore = cachestore or DiskCacheStore()
        chunks = chunks or {}
        data = _load(hash, cachestore, chunks)
        return cls(data[list(data.data_vars)[0]])

    @property
    def proportion(self) -> float:
        """Compute proportion of points with value 1.

        Returns
        -------
        float
            Proportion of points with value 1

        Raises
        ------
        NotImplementedError
            If instance does not contain binary data.
        """
        if not self.binary:
            raise NotImplementedError("proportion method is only implemented for binary fields")

        return self.data.sum().values.item() / self.data.count().values.item()

    def find_edges(self) -> MetDataArray:
        """Find edges of regions.

        Returns
        -------
        MetDataArray
            MetDataArray with a binary field, 1 on the edge of the regions,
            0 outside and inside the regions.

        Raises
        ------
        NotImplementedError
            If the instance is not binary.
        """
        if not self.binary:
            raise NotImplementedError("find_edges method is only implemented for binary fields")

        # edge detection algorithm using differentiation to reduce the areas to lines
        def _edges(da: xr.DataArray) -> xr.DataArray:
            lat_diff = da.differentiate("latitude")
            lon_diff = da.differentiate("longitude")
            diff = da.where((lat_diff != 0) | (lon_diff != 0), 0)

            # TODO: what is desired behavior here?
            # set boundaries to close contour regions
            diff[dict(longitude=0)] = da[dict(longitude=0)]
            diff[dict(longitude=-1)] = da[dict(longitude=-1)]
            diff[dict(latitude=0)] = da[dict(latitude=0)]
            diff[dict(latitude=-1)] = da[dict(latitude=-1)]

            return diff

        # load data into memory (required for value assignment in _edges()
        self.data.load()

        return MetDataArray(self.data.groupby("level").map(_edges), cachestore=self.cachestore)

    def to_polygon_feature(
        self,
        level: float | int | None = None,
        time: np.datetime64 | datetime | None = None,
        fill_value: float = 0.0,
        iso_value: float | None = None,
        min_area: float = 0.0,
        epsilon: float = 0.0,
        precision: int | None = None,
        interiors: bool = True,
        convex_hull: bool = False,
        include_altitude: bool = False,
        properties: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Create GeoJSON Feature artifact from spatial array on a single level and time slice.

        Computed polygons always contain an exterior linear ring as defined by the
        `GeoJSON Polygon specification <https://www.rfc-editor.org/rfc/rfc7946.html#section-3.1.6>`.
        Polygons may also contain interior linear rings (holes). This method does not support
        nesting beyond the GeoJSON specification. See the :mod:`pycontrails.core.polygon`
        for additional polygon support.

        .. versionchanged:: 0.25.12

            Previous implementation include several additional parameters which have
            been removed:

            - The ``approximate`` parameter
            - An ``path`` parameter to save output as JSON
            - Passing arbitrary kwargs to :func:`skimage.measure.find_contours`.

            New implementation includes new parameters previously lacking:

            - ``fill_value``
            - ``min_area``
            - ``include_altitude``

        .. versionchanged:: 0.38.0

            Change default value of ``epsilon`` from 0.15 to 0.

        .. versionchanged:: 0.41.0

            Convert continuous fields to binary fields before computing polygons.
            The parameters ``max_area`` and ``epsilon`` are now expressed in terms of
            longitude/latitude units instead of pixels.

        Parameters
        ----------
        level : float, optional
            Level slice to create polygons.
            If the "level" coordinate is length 1, then the single level slice will be selected
            automatically.
        time : datetime, optional
            Time slice to create polygons.
            If the "time" coordinate is length 1, then the single time slice will be selected
            automatically.
        fill_value : float, optional
            Value used for filling missing data and for padding the underlying data array.
            Expected to be less than the ``iso_value`` parameter. By default, 0.0.
        iso_value : float, optional
            Value in field to create iso-surface.
            Defaults to the average of the min and max value of the array. (This is the
            same convention as used by ``skimage``.)
        min_area : float, optional
            Minimum area of each polygon. Polygons with area less than ``min_area`` are
            not included in the output. The unit of this parameter is in longitude/latitude
            degrees squared. Set to 0 to omit any polygon filtering based on a minimal area
            conditional. By default, 0.0.
        epsilon : float, optional
            Control the extent to which the polygon is simplified. A value of 0 does not alter
            the geometry of the polygon. The unit of this parameter is in longitude/latitude
            degrees. By default, 0.0.
        precision : int, optional
            Number of decimal places to round coordinates to. If None, no rounding is performed.
        interiors : bool, optional
            If True, include interior linear rings (holes) in the output. True by default.
        convex_hull : bool, optional
            EXPERIMENTAL. If True, compute the convex hull of each polygon. Only implemented
            for depth=1. False by default. A warning is issued if the underlying algorithm
            fails to make valid polygons after computing the convex hull.
        include_altitude : bool, optional
            If True, include the array altitude [:math:`m`] as a z-coordinate in the
            `GeoJSON output <https://www.rfc-editor.org/rfc/rfc7946#section-3.1.1>`.
            False by default.
        properties : dict, optional
            Additional properties to include in the GeoJSON output. By default, None.

        Returns
        -------
        dict[str, Any]
            Python representation of GeoJSON Feature with MultiPolygon geometry.

        See Also
        --------
        :meth:`to_polyhedra`
        :func:`pycontrails.core.polygons.find_multipolygons`

        Examples
        --------
        >>> from pprint import pprint
        >>> from pycontrails.datalib.ecmwf import ERA5
        >>> era5 = ERA5("2022-03-01", variables="air_temperature", pressure_levels=250)
        >>> mda = era5.open_metdataset()["air_temperature"]
        >>> mda.shape
        (1440, 721, 1, 1)

        >>> pprint(mda.to_polygon_feature(iso_value=239.5, precision=2, epsilon=0.1))
        {'geometry': {'coordinates': [[[[167.88, -22.5],
                                        [167.75, -22.38],
                                        [167.62, -22.5],
                                        [167.75, -22.62],
                                        [167.88, -22.5]]],
                                      [[[43.38, -33.5],
                                        [43.5, -34.12],
                                        [43.62, -33.5],
                                        [43.5, -33.38],
                                        [43.38, -33.5]]]],
                      'type': 'MultiPolygon'},
         'properties': {},
         'type': 'Feature'}

        """
        if convex_hull and interiors:
            raise ValueError("Set 'interiors=False' to use the 'convex_hull' parameter.")

        arr, altitude = _extract_2d_arr_and_altitude(self, level, time)
        if not include_altitude:
            altitude = None  # this logic used below
        elif altitude is None:
            raise ValueError(
                "The parameter 'include_altitude' is True, but altitude is not "
                "found on MetDataArray instance. Either set altitude, or pass "
                "include_altitude=False."
            )

        np.nan_to_num(arr, copy=False, nan=fill_value)

        # default iso_value
        if iso_value is None:
            iso_value = (np.max(arr) + np.min(arr)) / 2
            warnings.warn(f"The 'iso_value' parameter was not specified. Using value: {iso_value}")

        # We'll get a nice error message if dependencies are not installed
        import pycontrails.core.polygon as polygon

        # Convert to nested lists of coordinates for GeoJSON representation
        longitude: npt.NDArray[np.float_] = self.variables["longitude"].values
        latitude: npt.NDArray[np.float_] = self.variables["latitude"].values

        mp = polygon.find_multipolygon(
            arr,
            threshold=iso_value,
            min_area=min_area,
            epsilon=epsilon,
            interiors=interiors,
            convex_hull=convex_hull,
            longitude=longitude,
            latitude=latitude,
            precision=precision,
        )

        return polygon.multipolygon_to_geojson(mp, altitude, properties)

    def to_polygon_feature_collection(
        self,
        time: np.datetime64 | datetime | None = None,
        fill_value: float = 0.0,
        iso_value: float | None = None,
        min_area: float = 0.0,
        epsilon: float = 0.0,
        precision: int | None = None,
        interiors: bool = True,
        convex_hull: bool = False,
        include_altitude: bool = False,
        properties: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Create GeoJSON FeatureCollection artifact from spatial array at time slice.

        See the :meth:`to_polygon_feature` method for a description of the parameters.

        Returns
        -------
        dict[str, Any]
            Python representation of GeoJSON FeatureCollection. This dictionary is
            comprised of individual GeoJON Features, one per :attr:`self.data["level"]`.
        """
        base_properties = properties or {}
        features = []
        for level in self.data["level"]:
            properties = base_properties.copy()
            properties.update(level=level.item())
            properties.update({f"level_{k}": v for k, v in self.data["level"].attrs.items()})

            feature = self.to_polygon_feature(
                level=level,
                time=time,
                fill_value=fill_value,
                iso_value=iso_value,
                min_area=min_area,
                epsilon=epsilon,
                precision=precision,
                interiors=interiors,
                convex_hull=convex_hull,
                include_altitude=include_altitude,
                properties=properties,
            )
            features.append(feature)

        return {
            "type": "FeatureCollection",
            "features": features,
        }

    @overload
    def to_polyhedra(
        self,
        *,
        time: datetime | None = ...,
        iso_value: float = ...,
        simplify_fraction: float = ...,
        return_type: Literal["geojson"],
        path: str | None = ...,
        altitude_scale: float = ...,
        output_vertex_normals: bool = ...,
        closed: bool = ...,
    ) -> dict:
        ...

    @overload
    def to_polyhedra(
        self,
        *,
        time: datetime | None = ...,
        iso_value: float = ...,
        simplify_fraction: float = ...,
        return_type: Literal["mesh"],
        path: str | None = ...,
        altitude_scale: float = ...,
        output_vertex_normals: bool = ...,
        closed: bool = ...,
    ) -> "o3d.geometry.TriangleMesh":
        ...

    def to_polyhedra(
        self,
        *,
        time: datetime | None = None,
        iso_value: float = 0.0,
        simplify_fraction: float = 1.0,
        return_type: str = "geojson",
        path: str | None = None,
        altitude_scale: float = 1.0,
        output_vertex_normals: bool = False,
        closed: bool = True,
    ) -> dict | "o3d.geometry.TriangleMesh":
        """Create a collection of polyhedra from spatial array corresponding to a single time slice.

        Parameters
        ----------
        time : datetime, optional
            Time slice to create mesh.
            If the "time" coordinate is length 1, then the single time slice will be selected
            automatically.
        iso_value : float, optional
            Value in field to create iso-surface. Defaults to 0.
        simplify_fraction : float, optional
            Apply `open3d` `simplify_quadric_decimation` method to simplify the polyhedra geometry.
            This parameter must be in the half-open interval (0.0, 1.0].
            Defaults to 1.0, corresponding to no reduction.
        return_type : str, optional
            Must be one of "geojson" or "mesh". Defaults to "geojson".
            If "geojson", this method returns a dictionary representation of a geojson MultiPolygon
            object whose polygons are polyhedra faces.
            If "mesh", this method returns an `open3d` `TriangleMesh` instance.
        path : str, optional
            Output geojson or mesh to file.
            If `return_type` is "mesh", see `Open3D File I/O for Mesh
            <http://www.open3d.org/docs/release/tutorial/geometry/file_io.html#Mesh>`_ for
            file type options.
        altitude_scale : float, optional
            Rescale the altitude dimension of the mesh, [:math:`m`]
        output_vertex_normals : bool, optional
            If ``path`` is defined, write out vertex normals.
            Defaults to False.
        closed : bool, optional
            If True, pad spatial array along all axes to ensure polyhedra are "closed".
            This flag often gives rise to cleaner visualizations. Defaults to True.

        Returns
        -------
        dict | :class:`o3d.geometry.TriangleMesh`
            Python representation of geojson object or `Open3D Triangle Mesh
            <http://www.open3d.org/docs/release/tutorial/geometry/mesh.html>`_ depending on the
            `return_type` parameter.

        Raises
        ------
        ModuleNotFoundError
            Method requires the `vis` optional dependencies
        ValueError
            If input parameters are invalid.

        See Also
        --------
        :meth:`to_polygons`
        `skimage.measure.marching_cubes <https://scikit-image.org/docs/dev/api/skimage.measure.html#skimage.measure.marching_cubes>`_

        Notes
        -----
        Uses the `scikit-image Marching Cubes  <https://scikit-image.org/docs/dev/auto_examples/edges/plot_marching_cubes.html>`_
        algorithm to reconstruct a surface from the point-cloud like arrays.
        """  # noqa: E501
        try:
            from skimage import measure
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                "This method requires the `skimage` module from scikit-learn, which can be "
                "installed using `pip install pycontrails[vis]`"
            ) from e

        try:
            import open3d as o3d
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                "This method requires the `open3d` module, which can be installed "
                "with `pip install pycontrails[open3d]` or `pip install open3d`."
            ) from e

        if len(self.data["level"]) == 1:
            raise ValueError(
                "Found single `level` coordinate in DataArray. This method requires at least two."
            )

        # select time
        if time is None and len(self.data["time"]) == 1:
            time = self.data["time"].values[0]

        if time is None:
            raise ValueError(
                "time input must be defined when the length of the time coordinates are > 1"
            )

        if simplify_fraction > 1 or simplify_fraction <= 0:
            raise ValueError("Parameter `simplify_fraction` must be in the interval (0, 1].")

        return_types = ["geojson", "mesh"]
        if return_type not in return_types:
            raise ValueError(f"Parameter `return_type` must be one of {', '.join(return_types)}")

        # 3d array of longitude, latitude, altitude values
        volume = self.data.sel(time=time).values

        # convert from array index back to coordinates
        longitude = self.variables["longitude"].values
        latitude = self.variables["latitude"].values
        altitude = units.pl_to_m(self.variables["level"].values)

        # Pad volume on all axes to close the volumes
        if closed:
            # pad values to domain
            longitude0 = longitude[0] - (longitude[1] - longitude[0])
            longitude1 = longitude[-1] + longitude[-1] - longitude[-2]
            longitude = np.pad(longitude, pad_width=1, constant_values=(longitude0, longitude1))

            latitude0 = latitude[0] - (latitude[1] - latitude[0])
            latitude1 = latitude[-1] + latitude[-1] - latitude[-2]
            latitude = np.pad(latitude, pad_width=1, constant_values=(latitude0, latitude1))

            altitude0 = altitude[0] - (altitude[1] - altitude[0])
            altitude1 = altitude[-1] + altitude[-1] - altitude[-2]
            altitude = np.pad(altitude, pad_width=1, constant_values=(altitude0, altitude1))

            # Pad along axes to ensure polygons are closed
            volume = np.pad(volume, pad_width=1, constant_values=iso_value)

        # Use marching cubes to obtain the surface mesh
        # Coordinates of verts are indexes of volume array
        verts, faces, normals, _ = measure.marching_cubes(
            volume, iso_value, allow_degenerate=False, gradient_direction="ascent"
        )

        # Convert from indexes to longitude, latitude, altitude values
        verts[:, 0] = longitude[verts[:, 0].astype(int)]
        verts[:, 1] = latitude[verts[:, 1].astype(int)]
        verts[:, 2] = altitude[verts[:, 2].astype(int)]

        # rescale altitude
        verts[:, 2] = verts[:, 2] * altitude_scale

        # create mesh in open3d
        mesh = o3d.geometry.TriangleMesh()
        mesh.vertices = o3d.utility.Vector3dVector(verts)
        mesh.triangles = o3d.utility.Vector3iVector(faces)
        mesh.vertex_normals = o3d.utility.Vector3dVector(normals)

        # simplify mesh according to sim
        if simplify_fraction < 1:
            target_n_triangles = int(faces.shape[0] * simplify_fraction)
            mesh = mesh.simplify_quadric_decimation(target_number_of_triangles=target_n_triangles)
            mesh.compute_vertex_normals()

        if path is not None:
            path = str(pathlib.Path(path).absolute())

        if return_type == "geojson":
            verts = np.round(
                np.asarray(mesh.vertices), decimals=4
            )  # rounding to reduce the size of resultant json arrays
            faces = np.asarray(mesh.triangles)

            # TODO: technically this is not valid GeoJSON because each polygon (triangle)
            # does not have the last element equal to the first (not a linear ring)
            # but it still works for now in Deck.GL
            coords = [[verts[face].tolist()] for face in faces]

            geojson = {
                "type": "Feature",
                "properties": {},
                "geometry": {"type": "MultiPolygon", "coordinates": coords},
            }

            if path is not None:
                if not path.endswith(".json"):
                    path += ".json"
                with open(path, "w") as file:
                    json.dump(geojson, file)
            return geojson

        if path is not None:
            o3d.io.write_triangle_mesh(
                path,
                mesh,
                write_ascii=False,
                compressed=True,
                write_vertex_normals=output_vertex_normals,
                write_vertex_colors=False,
                write_triangle_uvs=True,
                print_progress=False,
            )
        return mesh

    @overrides
    def broadcast_coords(self, name: str) -> xr.DataArray:
        da = xr.ones_like(self.data) * self.data[name]
        da.name = name

        return da

    @overrides
    def downselect(self, bbox: list[float]) -> MetDataArray:
        data = downselect(self.data, bbox)
        return MetDataArray(data, cachestore=self.cachestore)


def _is_wrapped(longitude: np.ndarray) -> bool:
    """Check if ``longitude`` covers ``[-180, 180]``."""
    return longitude[0] <= -180.0 and longitude[-1] >= 180.0


def _is_zarr(ds: xr.Dataset | xr.DataArray) -> bool:
    """Check if ``ds`` appears to be Zarr-based.

    Neither ``xarray`` nor ``dask`` readily expose such information, so
    implementation is very narrow and brittle.
    """
    if isinstance(ds, xr.Dataset):
        # Attempt 1: xarray binds the zarr close function to the instance
        # This gives us an indicator of the data is zarr based
        try:
            if ds._close.__func__.__qualname__ == "ZarrStore.close":  # type: ignore
                return True
        except AttributeError:
            pass

        # Grab the first data variable, get underlying DataArray
        ds = ds[next(iter(ds.data_vars))]

    # Attempt 2: Examine the dask graph
    darr = ds.variable._data  # dask array in some cases
    try:
        dask0 = darr.dask[next(iter(darr.dask))]  # first dask instruction
        return dask0.array.array.array.__class__.__name__ == "ZarrArrayWrapper"
    except AttributeError:
        return False


def shift_longitude(data: XArrayType) -> XArrayType:
    """Shift longitude values from [0, 360) to [-180, 180) domain.

    Sorts data by ascending longitude values.

    Parameters
    ----------
    data : XArrayType
        :class:`xr.Dataset` or :class:`xr.DataArray` with longitude dimension


    Returns
    -------
    XArrayType
        :class:`xr.Dataset` or :class:`xr.DataArray` with longitude values on [-180, 180).
    """
    return data.assign_coords(
        longitude=((data["longitude"].values + 180.0) % 360.0) - 180.0
    ).sortby("longitude", ascending=True)


def _wrap_longitude(data: XArrayType) -> XArrayType:
    """Wrap longitude grid coordinates.

    This function assumes the longitude dimension on ``data``:

    - is sorted
    - is contained in [-180, 180]

    These assumptions are checked by :class:`MetDataset` and :class`MetDataArray`
    constructors.

    .. versionchanged:: 0.26.0

        This function now ensures every value in the interval ``[-180, 180]``
        is covered by the longitude dimension of the returned object. See
        :meth:`MetDataset.is_wrapped` for more details.


    Parameters
    ----------
    data : XArrayType
        :class:`xr.Dataset` or :class:`xr.DataArray` with longitude dimension

    Returns
    -------
    XArrayType
        Copy of :class:`xr.Dataset` or :class:`xr.DataArray` with wrapped longitude values.

    Raises
    ------
    ValueError
        If longitude values are already wrapped.
    """
    if isinstance(data, xr.Dataset):
        lon = data.variables["longitude"].values
    else:
        lon = data.coords.variables["longitude"].values

    if _is_wrapped(lon):
        raise ValueError("Longitude values are already wrapped")

    lon0 = lon[0]
    lon1 = lon[-1]

    # Try to prevent something stupid
    if lon1 - lon0 < 330.0:
        warnings.warn("Wrapping longitude will create a large spatial gap of more than 30 degrees.")

    objs = [data]
    if lon0 > -180.0:  # if the lowest longitude is not low enough, duplicate highest
        dup1 = data.sel(longitude=[lon1]).assign_coords(longitude=[lon1 - 360.0])
        objs.insert(0, dup1)
    if lon1 < 180.0:  # if the highest longitude is not highest enough, duplicate lowest
        dup0 = data.sel(longitude=[lon0]).assign_coords(longitude=[lon0 + 360.0])
        objs.append(dup0)

    # Because we explicitly raise a ValueError if longitude already wrapped,
    # we know that len(objs) > 1, so the concatenation here is nontrivial
    wrapped = xr.concat(objs, dim="longitude")
    wrapped["longitude"] = wrapped["longitude"].astype(lon.dtype, copy=False)

    # If there is only one longitude chunk in parameter data, increment
    # data.chunks can be None, using getattr for extra safety
    # NOTE: This probably doesn't seem to play well with Zarr data ...
    # we don't want to be rechunking them.
    # Ideally we'd raise if data was Zarr-based
    chunks = getattr(data, "chunks", None) or {}
    chunks = dict(chunks)  # chunks can be frozen
    lon_chunks = chunks.get("longitude", tuple())
    if len(lon_chunks) == 1:
        chunks["longitude"] = (lon_chunks[0] + len(objs) - 1,)
        wrapped = wrapped.chunk(chunks)

    return wrapped


def _extract_2d_arr_and_altitude(
    mda: MetDataArray,
    level: float | int | None,
    time: np.datetime64 | datetime | None,
) -> tuple[np.ndarray, float | None]:
    """Extract underlying 2D array indexed by longitude and latitude.

    Parameters
    ----------
    mda : MetDataArray
        MetDataArray to extract from
    level : float | int | None
        Pressure level to slice at
    time : np.datetime64 | datetime
        Time to slice at

    Returns
    -------
    arr : np.ndarray
        Copy of 2D array at given level and time
    altitude : float | None
        Altitude of slice [:math:`m`]. None if "altitude" not found on ``mda``
        (ie, for surface level :class:`MetDataArray`).
    """
    # Determine level if not specified
    if level is None:
        level_coord = mda.variables["level"].values
        if len(level_coord) == 1:
            level = level_coord[0]
        else:
            raise ValueError(
                "Parameter 'level' must be defined when the length of the 'level' "
                "coordinates is not 1."
            )

    # Determine time if not specified
    if time is None:
        time_coord = mda.variables["time"].values
        if len(time_coord) == 1:
            time = time_coord[0]
        else:
            raise ValueError(
                "Parameter 'time' must be defined when the length of the 'time' "
                "coordinates is not 1."
            )

    da = mda.data.sel(level=level, time=time)
    arr = da.values.copy()
    if arr.ndim != 2:
        raise RuntimeError("Malformed data array")

    try:
        altitude = da["altitude"].values.item()  # item not implemented on dask arrays
    except KeyError:
        altitude = None
    else:
        altitude = round(altitude)

    return arr, altitude


def downselect(data: XArrayType, bbox: list[float]) -> XArrayType:
    """Downselect :class:`xr.Dataset` or :class:`xr.DataArray` with spatial bounding box.

    Parameters
    ----------
    data : XArrayType
        xr.Dataset or xr.DataArray to downselect
    bbox : list[float]
        List of coordinates defining a spatial bounding box in WGS84 coordinates.
        For 2D queries, list is [west, south, east, north].
        For 3D queries, list is [west, south, min-level, east, north, max-level]
        with level defined in [:math:`hPa`].

    Returns
    -------
    XArrayType
        Downselected xr.Dataset or xr.DataArray

    Raises
    ------
    ValueError
        If parameter `bbox` has wrong length.
    """
    if len(bbox) == 4:
        west, south, east, north = bbox
        level_min = -np.inf
        level_max = np.inf

    elif len(bbox) == 6:
        west, south, level_min, east, north, level_max = bbox

    else:
        raise ValueError(
            f"bbox {bbox} is not length 4 [west, south, east, north] "
            "or length 6 [west, south, min-level, east, north, max-level]"
        )

    cond = (
        (data["latitude"] >= south)
        & (data["latitude"] <= north)
        & (data["level"] >= level_min)
        & (data["level"] <= level_max)
    )

    # wrapping longitude
    if west <= east:
        cond = cond & (data["longitude"] >= west) & (data["longitude"] <= east)
    else:
        cond = cond & ((data["longitude"] >= west) | (data["longitude"] <= east))

    return data.where(cond, drop=True)


def standardize_variables(ds: DatasetType, variables: Iterable[MetVariable]) -> DatasetType:
    """Rename all variables in dataset from short name to standard name.

    This function does not change any variables in ``ds`` that are not found in ``variables``.

    When there are multiple variables with the same short name, the last one is used.

    Parameters
    ----------
    ds : DatasetType
        An :class:`xr.Dataset` or :class:`MetDataset`. When a :class:`MetDataset` is
        passed, the underlying :class:`xr.Dataset` is modified in place.
    variables : Iterable[MetVariable]
        Data source variables

    Returns
    -------
    DatasetType
        Dataset with variables renamed to standard names
    """
    if isinstance(ds, xr.Dataset):
        return _standardize_variables(ds, variables)

    ds.data = _standardize_variables(ds.data, variables)
    return ds


def _standardize_variables(ds: xr.Dataset, variables: Iterable[MetVariable]) -> xr.Dataset:
    variables_dict: dict[Hashable, str] = {v.short_name: v.standard_name for v in variables}
    name_dict = {var: variables_dict[var] for var in ds.data_vars if var in variables_dict}
    return ds.rename(name_dict)


def originates_from_ecmwf(met: MetDataset | MetDataArray) -> bool:
    """Check if data appears to have originated from an ECMWF source.

    .. versionadded:: 0.27.0

        Experimental. Implementation is brittle.

    Parameters
    ----------
    met : MetDataset | MetDataArray
        Dataset or array to inspect.

    Returns
    -------
    bool
        True if data appears to be derived from an ECMWF source.

    See Also
    --------
    - :class:`ERA5`
    - :class:`HRES`

    """
    return met.attrs.get("met_source") in ("ERA5", "HRES") or "ecmwf" in met.attrs.get(
        "history", ""
    )


def _load(hash: str, cachestore: CacheStore, chunks: dict[str, int]) -> xr.Dataset:
    """Load xarray data from hash.

    Parameters
    ----------
    hash : str
        Description
    cachestore : CacheStore
        Description
    chunks : dict[str, int]
        Description

    Returns
    -------
    xr.Dataset
        Description
    """
    disk_path = cachestore.get(f"{hash}*.nc")
    return xr.open_mfdataset(disk_path, chunks=chunks)



"""Test models and methods pertaining to ISSR, SAC, and PCR."""

from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from pycontrails import Flight, JetA, MetDataArray, MetDataset
from pycontrails.core.met import originates_from_ecmwf
from pycontrails.models import sac
from pycontrails.models.humidity_scaling import (
    ConstantHumidityScaling,
    ExponentialBoostHumidityScaling,
)
from pycontrails.models.issr import ISSR
from pycontrails.models.pcr import PCR
from pycontrails.models.sac import SAC
from pycontrails.physics import constants, thermo, units
from tests import _deprecated

# TODO: Include tests for all simple models here
# 1) Downselect met results in copy
# 2) Check downselect met bounds
# 4) Default params for SAC, etc


def test_ISSR_met_source(met_issr: MetDataset) -> None:
    """Test ISSR model evaluated on a MetDataset source."""

    with pytest.warns(UserWarning, match="originated from ECMWF"):
        ISSR(met=met_issr)

    model = ISSR(met=met_issr, humidity_scaling=ExponentialBoostHumidityScaling())
    out = model.eval()

    assert isinstance(out, MetDataArray)
    assert isinstance(out.data, xr.DataArray)
    assert isinstance(model.met, MetDataset)
    assert out.name == "issr"

    assert out.attrs["humidity_scaling_name"] == "exponential_boost"
    assert out.attrs["humidity_scaling_formula"] == "rhi -> (rhi / rhi_adj) ^ rhi_boost_exponent"
    assert out.attrs["humidity_scaling_rhi_adj"] == 0.97
    assert out.attrs["humidity_scaling_rhi_boost_exponent"] == 1.7
    assert out.attrs["humidity_scaling_clip_upper"] == 1.7

    assert out.attrs["description"] == "Ice super-saturated regions"
    assert "pycontrails_version" in out.attrs
    assert "met_source" in out.attrs

    # Pin some counts
    vals, counts = np.unique(out.data, return_counts=True)
    np.testing.assert_array_equal(vals, [0, 1])
    np.testing.assert_array_equal(counts, [563, 157])

    met_copy = met_issr.copy()
    del met_copy.data["air_temperature"]
    out = model.eval(source=met_copy)

    # Important: if specific_humidity included in model input, it will NOT
    # undergo humidity scaling. This is why the counts below are lower.
    assert "humidity_scaling" not in out.attrs
    assert isinstance(out, MetDataArray)

    # Pin some counts
    vals, counts = np.unique(out.data, return_counts=True)
    np.testing.assert_array_equal(vals, [0, 1])
    np.testing.assert_array_equal(counts, [587, 133])


def test_SAC_met_source(met_issr: MetDataset) -> None:
    """Test the SAC model evaluated in a MetDataset source."""

    with pytest.warns(UserWarning, match="originated from ECMWF"):
        SAC(met=met_issr)

    model = SAC(met_issr, humidity_scaling=ConstantHumidityScaling())
    out1 = model.eval()

    assert isinstance(out1, MetDataArray)
    assert isinstance(out1.data, xr.DataArray)
    assert isinstance(model.met, MetDataset)
    assert out1.name == "sac"

    # Pin some counts
    vals, counts = np.unique(out1.data, return_counts=True)
    np.testing.assert_array_equal(vals, [0, 1])
    np.testing.assert_array_equal(counts, [351, 369])

    met_copy = met_issr.copy()
    del met_copy.data["air_temperature"]
    met_copy.data["engine_efficiency"] = 0.35
    out2 = model.eval(source=met_copy)

    assert isinstance(out2, MetDataArray)
    assert out2.name == "sac"
    assert np.all(model.source.data["engine_efficiency"].values == 0.35)

    # Pin some counts
    # Values are now different because humidity was not scaled
    vals, counts = np.unique(out2.data, return_counts=True)
    np.testing.assert_array_equal(vals, [0, 1])
    np.testing.assert_array_equal(counts, [340, 380])


def test_SAC_with_nan(met_issr: MetDataset) -> None:
    """Check that NaN values persist after SAC calculation."""
    met = met_issr.copy()
    met.data.load()
    met.data["specific_humidity"][4, 3, 2, 1] = np.nan

    model = SAC(met, humidity_scaling=ExponentialBoostHumidityScaling())
    out = model.eval()
    np.testing.assert_array_equal(np.nonzero(np.isnan(out.values)), [[4], [3], [2], [1]])

    actual = np.unique(out.data)
    expected = np.array([0.0, 1.0, np.nan])
    np.testing.assert_array_equal(actual, expected)


def test_PCR_grid(met_issr: MetDataset) -> None:
    """Test PCRGrid model."""

    with pytest.warns(UserWarning, match="originated from ECMWF"):
        PCR(met=met_issr)

    model = PCR(met_issr, humidity_scaling=ConstantHumidityScaling())
    out1 = model.eval()

    assert isinstance(out1, MetDataArray)
    assert isinstance(out1.data, xr.DataArray)
    assert isinstance(model.met, MetDataset)
    assert out1.name == "pcr"

    assert out1.attrs.pop("pycontrails_version")
    assert out1.attrs == {
        "description": "Persistent contrail regions",
        "humidity_scaling_name": "constant_scale",
        "humidity_scaling_formula": "rhi -> rhi / rhi_adj",
        "humidity_scaling_rhi_adj": 0.97,
        "met_source": "ERA5",
        "engine_efficiency": 0.3,
    }

    # Pin some counts
    vals, counts = np.unique(out1.data, return_counts=True)
    np.testing.assert_array_equal(vals, [0, 1])
    np.testing.assert_array_equal(counts, [607, 113])

    met_copy = met_issr.copy()
    del met_copy.data["air_temperature"]
    met_copy.data["engine_efficiency"] = 0.35
    out2 = model.eval(source=met_copy)

    assert isinstance(out2, MetDataArray)
    assert out2.name == "pcr"
    assert np.all(model.source.data["engine_efficiency"].values == 0.35)

    assert out2.attrs.pop("pycontrails_version")
    assert out2.attrs == {
        "description": "Persistent contrail regions",
        "met_source": "ERA5",
    }

    # Pin some counts
    vals, counts = np.unique(out2.data, return_counts=True)
    np.testing.assert_array_equal(vals, [0, 1])
    np.testing.assert_array_equal(counts, [622, 98])


def test_era5_as_expected(met_era5_fake: MetDataset) -> None:
    """Test ERA5 data with all models."""

    # created with exactly 20% is ISSR
    assert not originates_from_ecmwf(met_era5_fake)
    issr = ISSR(met=met_era5_fake).eval()
    assert isinstance(issr, MetDataArray)
    assert issr.proportion == 0.2

    # ISSR exist exactly when longitude is a multiple of 5
    # demonstrating this behavior on a few arbitrary selections
    assert np.all(issr.data.sel(longitude=123).values == 0)
    assert np.all(issr.data.sel(longitude=124).values == 0)
    assert np.all(issr.data.sel(longitude=125).values == 1)
    assert np.all(issr.data.sel(longitude=126).values == 0)
    assert np.all(issr.data.sel(longitude=127).values == 0)

    # ISSR and PCR do not completely agree for this data
    # In particular, on levels 150, 175, and 200, the two diverge
    pcr = PCR(met_era5_fake).eval()
    assert isinstance(pcr, MetDataArray)
    pcr_low = pcr.data.isel(level=slice(3, None))
    issr_low = issr.data.isel(level=slice(3, None))
    xr.testing.assert_equal(pcr_low, issr_low)

    # But pcr is identically 0 on the top three levels
    pcr_high = pcr.data.isel(level=slice(0, 3))
    assert (pcr_high == 0).all()


def test_ISSR_flight(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Test ISSR model."""
    out = ISSR(met=met_era5_fake).eval(source=flight_fake)

    assert isinstance(out, Flight)
    assert np.nansum(out["issr"]) == 73.0


def test_ISSR_flight_high_altitude(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Test ISSR model with high altitudes."""
    # modifying fixture
    fl = flight_fake.copy()
    fl["altitude"] *= 2.5

    with pytest.warns(UserWarning, match="Flight altitude is high"):
        with pytest.raises(ValueError, match="One of the requested xi is out of bounds"):
            ISSR(met=met_era5_fake, interpolation_bounds_error=True).eval(source=fl)

    with pytest.warns(UserWarning, match="Flight altitude is high"):
        fl = ISSR(met=met_era5_fake).eval(source=fl)

    # waypoints above highest-altitude level in era5 get filled with NaN
    assert isinstance(fl, Flight)
    low_level_waypoints = fl.level < met_era5_fake.data["level"].min().item()
    assert np.isnan(fl["issr"][low_level_waypoints]).all()
    assert ~np.isnan(fl["issr"][~low_level_waypoints]).all()


@pytest.mark.parametrize("interpolation_use_indices", [True, False])
def test_ISSR_flight_outside_met(
    met_era5_fake: MetDataset, flight_fake: Flight, interpolation_use_indices: bool
) -> None:
    """Test ISSR model with flight data outside bounds of met data.

    The flight trajectory leaves the met domain in two places:

    - the initial flight waypoints are at low altitude (below the max pressure level in met)
    - the later flight waypoints have times beyond the max time in met
    """

    # Alter the flight time
    fl = flight_fake.copy()
    with pytest.warns(UserWarning, match="Overwriting data in key `time`."):
        fl["time"] += pd.Timedelta("1H")

    model = ISSR(met_era5_fake, interpolation_use_indices=interpolation_use_indices)

    with pytest.raises(ValueError, match="One of the requested xi is out of bounds in dimension 2"):
        model.eval(source=fl, interpolation_bounds_error=True)

    # Evaluate the model
    fl = model.eval(source=fl, interpolation_bounds_error=False)
    assert isinstance(fl, Flight)
    issr = fl["issr"]
    assert issr.dtype == np.float64

    # Waypoints with timestamp beyond the max time are nan-filled
    late_waypoints = fl["time"] > met_era5_fake["time"].values[-1]
    assert np.sum(late_waypoints) == 130
    assert np.isnan(issr[late_waypoints]).all()

    # Waypoints with low altitude are also nan-filled
    low_waypoints = fl.altitude < met_era5_fake["altitude"].values.min()
    assert np.sum(low_waypoints) == 257
    assert np.isnan(fl["issr"][late_waypoints]).all()

    # Anything remaining is non-nan
    remaining = ~late_waypoints & ~low_waypoints
    assert np.sum(remaining) == 113
    assert np.all(np.isfinite(issr[remaining]))


def test_ISSR_flight_span_antimeridian(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Test ISSR model with flight data spanning anti-meridian."""
    # modifying fixture
    met_era5_fake_wrapped = MetDataset(met_era5_fake.data, wrap_longitude=True)
    fl = flight_fake.copy()
    n_waypoints = len(fl)
    assert n_waypoints % 2 == 0  # needed below
    longitude = np.concatenate(
        [
            np.linspace(150, 180, n_waypoints // 2, endpoint=False),
            np.linspace(-180, -140, n_waypoints // 2),
        ]
    )
    altitude = np.full(shape=fl.shape, fill_value=10000)
    fl.update(longitude=longitude, altitude=altitude)

    fl = ISSR(met_era5_fake_wrapped).eval(source=fl)
    assert isinstance(fl, Flight)
    assert np.all(~np.isnan(fl["issr"]))

    # assert output
    assert fl["issr"].sum() == 158.0


def test_ISSR_flight_fill_value(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Test ISSR model with nan fill value."""

    # Waypoints below lowest-altitude level in era5 get filled with nan
    low_altitude_waypoints = flight_fake["altitude"] < met_era5_fake.data["altitude"].min().item()

    # null result (default fill value)
    out = ISSR(met_era5_fake).eval(source=flight_fake)
    assert np.isnan(out["issr"][low_altitude_waypoints]).all()
    assert np.isfinite(out["issr"][~low_altitude_waypoints]).all()

    # When we pass in a custom fill_value here, we no longer encounter nan
    # keep the interpolation fill value reasonable here; else some of our
    # thermo functions will encounter divide by zero issues
    out2 = ISSR(met_era5_fake, interpolation_fill_value=180).eval(source=flight_fake)
    assert np.isfinite(out2["issr"]).all()


def test_ISSR_flight_interp(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Test ISSR model with interpolation methods."""

    # test fake method
    with pytest.raises(ValueError, match="Method 'fake' is not defined"):
        _ = ISSR(met_era5_fake, interpolation_method="fake").eval(source=flight_fake)

    # test bounds error
    with pytest.raises(ValueError, match="xi is out of bounds"):
        _ = ISSR(met_era5_fake, interpolation_bounds_error=True).eval(source=flight_fake)

    # default interpolation method is linear
    out2 = ISSR(met_era5_fake).eval(source=flight_fake)
    out3 = ISSR(met_era5_fake, interpolation_method="linear").eval(source=flight_fake)

    assert isinstance(out2, Flight)
    assert isinstance(out3, Flight)

    # linear fill is default, so these must be equal (outside nan)
    assert np.all(out2["issr"][~np.isnan(out2["issr"])] == out3["issr"][~np.isnan(out3["issr"])])

    # nearest interpolation without extrapolation
    out4 = ISSR(met_era5_fake, interpolation_method="nearest").eval(source=flight_fake)
    # nearest interpolation with extrapolation
    out5 = ISSR(met_era5_fake, interpolation_method="nearest", interpolation_fill_value=None).eval(
        source=flight_fake
    )

    assert isinstance(out4, Flight)
    assert isinstance(out5, Flight)

    # out4 has missing values, out5 has no missing values
    assert np.isnan(out4["issr"]).any()
    assert not np.isnan(out5["issr"]).any()

    # AND, because fl4 is grabbing the closest era5 value
    # it contains positive ISSR waypoints at very low altitudes
    assert out5["issr"][0] == 1

    # expect an issr value exactly when flight waypoint
    # longitude value is within 1/2 of a multiple of 5
    expect_issr = np.abs((flight_fake["longitude"] + 2.5) % 5 - 2.5) < 1 / 2
    actual_issr = out5["issr"].astype(bool)
    assert np.all(expect_issr == actual_issr)


def test_ISSR_flight_incomplete_era5(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Test ISSR flight with incomplete met data."""

    da = met_era5_fake["air_temperature"].data
    ds = da.to_dataset()
    mds = MetDataset(ds)
    with pytest.raises(KeyError, match="Dataset does not contain variable `specific_humidity`"):
        ISSR(mds)


def test_ISSR_flight_missing_waypoint(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Test ISSR flight with missing waypoints met data."""

    # modifying fl fixture to avoid NaN caused by low altitude waypoints
    fl = flight_fake.copy()
    fl.update(altitude=np.full(shape=fl.shape, fill_value=10000.0))

    # and adding some additional NaN
    fl["altitude"][200] = np.nan
    fl["longitude"][210] = np.nan
    fl["latitude"][212] = np.nan
    fl["time"][217] = np.datetime64("NaT")

    with pytest.warns(UserWarning, match="NaT times"):
        out2 = ISSR(met_era5_fake).eval(source=fl)

    # expect issr column to have exactly at waypoints with NaN
    for idx in [200, 210, 212, 217]:
        assert np.isnan(out2["issr"][idx])


def test_SAC_flight(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Test SAC flight model."""

    fl = flight_fake.copy()

    prev_sac_sum = float("inf")
    for altitude, sac_sum in [(9000, 160), (10000, 150), (11000, 139), (12000, 128), (13000, 0)]:
        # modifying flight_fake fixture by setting constant altitude
        fl.update(altitude=np.full(shape=fl.shape, fill_value=altitude))
        fl.attrs["engine_efficiency"] = 0.4

        model = SAC(met_era5_fake, humidity_scaling=ExponentialBoostHumidityScaling())
        out = model.eval(source=fl)
        assert isinstance(out, Flight)
        assert np.nansum(out["sac"]) == sac_sum

        # NOTE: as altitudes go up, we get fewer SAC
        # this is caused by the `air_pressure` term in the function `G` in formula.py
        cur_sac_sum = np.nansum(out["sac"])
        assert cur_sac_sum < prev_sac_sum
        prev_sac_sum = cur_sac_sum


def test_SAC_flight_with_nan(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Check that NaN persist in SAC calculation."""

    # Original flight has no NaN
    for key in flight_fake:
        assert not np.isnan(flight_fake[key]).any()

    # So we put some in
    fl = flight_fake.copy()
    fl["longitude"][5] = np.nan

    # But, the flight is out of bounds of the met data
    with pytest.raises(ValueError, match="bounds"):
        SAC(met_era5_fake, interpolation_bounds_error=True).eval(source=fl)

    # So modify the level to keep flight in the met bounds
    fl.update(altitude=np.full(shape=fl.shape, fill_value=units.pl_to_m(275)))

    # And confirm that the NaN at position 5 persists in the new SAC column
    # Note that "nan" is considered out of bounds, so we need to set bounds error to False
    out = SAC(met_era5_fake, interpolation_bounds_error=False).eval(source=fl)
    sac = out["sac"]
    actual = np.isnan(sac)
    expected = np.zeros_like(sac).astype(bool)
    expected[5] = True
    np.testing.assert_array_equal(actual, expected)


def test_PCR_flight(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Test PCR Flight model."""

    model = PCR(met_era5_fake, humidity_scaling=ExponentialBoostHumidityScaling())

    out1 = model.eval(source=flight_fake)
    assert isinstance(out1, Flight)
    assert "engine_efficiency" in out1.attrs  # params transferred from model
    assert np.nansum(out1["pcr"]) == 56

    # override engine efficiency via attributes
    flight_fake.attrs["engine_efficiency"] = 0.4
    out2 = model.eval(source=flight_fake)

    assert isinstance(out2, Flight)
    assert out2.attrs["engine_efficiency"] == 0.4
    assert np.nansum(out2["pcr"]) == 71


@pytest.mark.parametrize("step", [1e-3, 1e-6])
def test_e_sat_liquid_prime(step: float) -> None:
    """Confirm `sac._e_sat_liquid_prime` formula agrees with secant approximation.

    This test as well as the `sac._e_sat_liquid_prime` function will need to be
    updated if `thermo.e_sat_liquid` changes.
    """
    T0 = np.linspace(-60, -20, 200) - constants.absolute_zero
    assert np.all(T0 > 0)

    T1 = T0 + step * np.ones_like(T0)

    e_sat0 = thermo.e_sat_liquid(T0)
    e_sat1 = thermo.e_sat_liquid(T1)

    secant_slope = (e_sat1 - e_sat0) / (T1 - T0)
    tangent_slope = sac._e_sat_liquid_prime(T0)

    error = np.abs(secant_slope - tangent_slope)
    assert np.all(error < step)  # approximation is O(step)


def test_T_critical_sac_implementation() -> None:
    """Show agreement between current scipy-based approach with depreciated version."""
    rng = np.random.default_rng(987654)
    size = 10000
    p = rng.uniform(20000, 40000, size)
    q = rng.uniform(0, 0.0001, size)
    t = rng.uniform(-70, -10, size) - constants.absolute_zero
    eta = rng.uniform(0.2, 0.5, size)
    rh = thermo.rh(q, t, p)

    jetA = JetA()
    ei_h2o = jetA.ei_h2o
    q_fuel = jetA.q_fuel

    G = sac.slope_mixing_line(q, p, eta, ei_h2o, q_fuel)
    t_lm = sac.T_sat_liquid(G)
    t_crit1 = sac.T_critical_sac(t_lm, rh, G)

    # We get a bunch of warnings here
    # We acknowledge them and continue
    with pytest.warns(RuntimeWarning):
        t_crit2 = _deprecated.T_critical_sac(t, q, p, eta)  # type: ignore[arg-type]

    assert np.all(np.isfinite(t_crit1))
    assert np.all(np.isfinite(t_crit2))

    # Close agreement
    np.testing.assert_allclose(t_crit1, t_crit2, atol=0.15)


def test_T_sat_liquid() -> None:
    """Confirm close agreement between `T_sat_liquid` and `T_sat_liquid_high_accuracy`."""
    rng = np.random.default_rng(737)
    size = 10000
    G = rng.uniform(low=1, high=4, size=size)
    t1 = sac.T_sat_liquid(G)
    t2 = sac.T_sat_liquid_high_accuracy(G)

    # Close agreement
    np.testing.assert_allclose(t1, t2, atol=0.05)


def test_T_critical_sac_schumann() -> None:
    """Confirm Schumann's statements from his 1996 paper."""
    rng = np.random.default_rng(737)
    size = 10000
    G = rng.uniform(low=1, high=4, size=size)

    # Intentionally going above 1 for rh here; these supersaturated values
    # get clipped in sac.T_critical_sac
    rh = rng.uniform(low=0, high=1.01, size=size)
    t_LM = sac.T_sat_liquid(G)
    t_LC = sac.T_critical_sac(t_LM, rh, G)

    # Critical temperature is always below tangent line temperature value
    assert np.all(t_LC <= t_LM)
    # Equality when U >= 1
    idx = np.nonzero(rh >= 1)
    np.testing.assert_array_equal(t_LC[idx], t_LM[idx])

    # For U = 0, equation (11) gives explicit answer
    t_LC_U0 = sac.T_critical_sac(t_LM, np.zeros_like(rh), G)
    eqn11 = t_LM - thermo.e_sat_liquid(t_LM) / G
    np.testing.assert_array_equal(t_LC_U0, eqn11)


@pytest.mark.parametrize("in_bounds", [False, True])
def test_issr_source_grid_different_resolution(met_issr: MetDataset, in_bounds: bool):
    """Confirm the ISSR model can handle gridded source with different resolution from met."""
    model = ISSR(met=met_issr, humidity_scaling=ConstantHumidityScaling(rhi_adj=0.95))

    # Create source
    level = [240, 250, 271]
    if not in_bounds:
        level[0] = 221

    source = MetDataset.from_coords(
        longitude=[43, 44, 45],
        latitude=[-80, -55, 13],
        level=level,
        time=np.datetime64("2019-05-31 05:00:00"),
    )

    if in_bounds:
        out = model.eval(source, interpolation_bounds_error=True)
        assert np.all(np.isfinite(out.data))
        return

    with pytest.raises(ValueError, match="221"):
        model.eval(source, interpolation_bounds_error=True)



"""Test `Cocip`."""

from __future__ import annotations

import json
import pathlib
import time as pythontime

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from pycontrails import Aircraft, Flight, MetDataset
from pycontrails.core.met import originates_from_ecmwf
from pycontrails.models import humidity_scaling as hs
from pycontrails.models.aircraft_performance import AircraftPerformance
from pycontrails.models.cocip import (
    Cocip,
    CocipFlightParams,
    contrail_properties,
    radiative_heating,
)
from pycontrails.models.cocip.output import grid_cirrus
from pycontrails.models.humidity_scaling import (
    ExponentialBoostHumidityScaling,
    ExponentialBoostLatitudeCorrectionHumidityScaling,
)

from .conftest import get_static_path


@pytest.fixture
def met(met_cocip1: MetDataset) -> MetDataset:
    """Rename fixture `met_cocip1` from conftest."""
    return met_cocip1


@pytest.fixture
def rad(rad_cocip1: MetDataset) -> MetDataset:
    """Rename fixture `rad_cocip1` from conftest."""
    return rad_cocip1


@pytest.fixture
def fl(flight_cocip1: Flight) -> Flight:
    """Rename fixture `cocip_fl` from conftest."""
    return flight_cocip1


@pytest.fixture
def cocip_no_ef(fl: Flight, met: MetDataset, rad: MetDataset) -> Cocip:
    """Return `Cocip` instance evaluated on modified `fl`."""
    fl.update(longitude=np.linspace(-29, -32, 20))
    fl.update(latitude=np.linspace(54, 55, 20))
    fl.update(altitude=np.full(20, 10000))

    # set all radiative forcing to 0
    rad2 = rad.copy()
    rad2.data["top_net_solar_radiation"] = xr.zeros_like(rad2.data["top_net_solar_radiation"])
    rad2.data["top_net_thermal_radiation"] = xr.zeros_like(rad2.data["top_net_thermal_radiation"])

    # run - will not find any persistent contrails
    params = {
        "max_age": np.timedelta64(3, "h"),
        "process_emissions": False,
        "humidity_scaling": ExponentialBoostHumidityScaling(),
    }
    cocip = Cocip(met.copy(), rad=rad2, params=params)
    cocip.eval(source=fl)

    return cocip


@pytest.fixture
def cocip_persistent(fl: Flight, met: MetDataset, rad: MetDataset) -> Cocip:
    """Return `Cocip` instance evaluated on modified `fl`."""
    fl.update(longitude=np.linspace(-29, -32, 20))
    fl.update(latitude=np.linspace(56, 57, 20))
    fl.update(altitude=np.linspace(10900, 10900, 20))

    # Use a non-default met_time_buffer for backwards compatibility with original
    # pinned test data. This only affects the test_grid_cirrus test below. Flight
    # waypoint data remains unchanged.
    params = {
        "max_age": np.timedelta64(3, "h"),
        "process_emissions": False,
        "verbose_outputs": True,
        "met_time_buffer": (np.timedelta64(0, "h"), np.timedelta64(1, "h")),
        "humidity_scaling": ExponentialBoostHumidityScaling(),
    }
    cocip = Cocip(met.copy(), rad=rad.copy(), params=params)

    # Eventually the advected waypoints will blow out of bounds
    # Specifically, there is a contrail at 4:30 with latitude larger than 59
    # We acknowledge this here
    with pytest.warns(UserWarning, match="outside of the met interpolation grid"):
        cocip.eval(source=fl)

    return cocip


@pytest.fixture
def cocip_persistent2(
    flight_cocip2: Flight, met_cocip2: MetDataset, rad_cocip2: MetDataset
) -> Cocip:
    """Return ``Cocip`` instance evaluated on modified ``flight_cocip2``."""

    # Use a non-default met_time_buffer for backwards compatibility with original
    # pinned test data. This only affects the test_grid_cirrus test below. Flight
    # waypoint data remains unchanged.
    params = {
        "max_age": np.timedelta64(5, "h"),
        "process_emissions": False,
        "verbose_outputs": True,
        "interpolation_bounds_error": True,
        "humidity_scaling": ExponentialBoostHumidityScaling(),
    }
    cocip = Cocip(met=met_cocip2.copy(), rad=rad_cocip2.copy(), params=params)
    cocip.eval(source=flight_cocip2)

    return cocip


def test_init_model(met: MetDataset, rad: MetDataset) -> None:
    """Test `Cocip` initialization."""
    # require rad
    with pytest.raises(TypeError, match="positional argument: 'rad'"):
        Cocip(met=met)

    # warn if humidity scaling not specified
    with pytest.warns(UserWarning, match="originated from ECMWF"):
        Cocip(met=met, rad=rad)

    # init
    cocip = Cocip(met=met, rad=rad, humidity_scaling=ExponentialBoostHumidityScaling())

    # default params
    assert isinstance(cocip.params, dict)
    cp = CocipFlightParams().as_dict()
    for key in cp:
        if key == "humidity_scaling":
            continue
        v1 = cocip.params[key]
        v2 = cp[key]
        try:
            assert v1 == v2 or (np.isnan(v1) and np.isnan(v2))
        except ValueError:  # np.ndarray
            np.testing.assert_array_equal(v1, v2)

    # model hash
    assert isinstance(cocip.hash, str)

    # inputs
    assert cocip.met is met
    assert cocip.rad is rad

    # outputs
    assert cocip.contrail is None
    assert cocip.contrail_dataset is None


def test_cocip_processes_met(met: MetDataset, rad: MetDataset) -> None:
    """Check that Cocip seamlessly processes met data on init."""

    # met starts without tau_cirrus
    assert "tau_cirrus" not in met
    q = met["specific_humidity"].values.copy()

    cocip = Cocip(met=met, rad=rad, humidity_scaling=ExponentialBoostHumidityScaling())

    # tau_cirrus added to met in __init__
    assert cocip.met is met
    assert "tau_cirrus" in met

    # specific humidity no longer modified on met
    np.testing.assert_array_equal(q, met["specific_humidity"].values)

    # There is not longer any issue in instantiating another model
    another_cocip = Cocip(met=met, rad=rad, humidity_scaling=ExponentialBoostHumidityScaling())
    assert another_cocip.met is met


def test_cocip_processes_rad(met: MetDataset, rad: MetDataset) -> None:
    """Check that Cocip seamlessly processes rad data on init."""

    # A little HACK to avoid the warning
    assert originates_from_ecmwf(met)
    del met.attrs["history"]
    assert not originates_from_ecmwf(met)

    # rad starts without time shifting
    t0 = pd.date_range("2019", freq="1H", periods=13)
    t1 = t0 + Cocip.default_params.shift_radiation_time
    np.testing.assert_array_equal(rad["time"].values, t0)
    assert "_pycontrails_modified" not in rad["time"].attrs
    cocip = Cocip(met=met, rad=rad)

    # time shifted in __init__
    assert cocip.rad is rad
    assert rad["time"].attrs["_pycontrails_modified"]
    assert rad["time"].attrs["shift_radiation_time"] == "-30 minutes"
    np.testing.assert_array_equal(rad["time"].values, t1)

    # should run again no problem
    another_cocip = Cocip(met=met, rad=rad)
    assert another_cocip.rad is rad

    # radiation not shifted a second time
    np.testing.assert_array_equal(rad["time"].values, t1)

    # drop a required rad variable, get an error
    rad.data = rad.data.drop_vars(["top_net_thermal_radiation"])
    with pytest.raises(KeyError, match="Dataset does not contain variable"):
        Cocip(met=met, rad=rad)


def test_cocip_processes_rad_with_warnings(met: MetDataset, rad: MetDataset) -> None:
    """Check that Cocip issues warnings for when rad data is different from expected."""

    del met.attrs["history"]  # see test_cocip_processes_rad

    original_time = rad["time"].values.copy()

    with pytest.warns(UserWarning, match="Shifting radiation time dimension by unexpected"):
        Cocip(met, rad=rad, shift_radiation_time=np.timedelta64(20, "m"))

    assert "shift_radiation_time" in rad["time"].attrs
    assert rad["time"].attrs["shift_radiation_time"] == "20 minutes"
    np.testing.assert_array_equal(original_time + np.timedelta64(20, "m"), rad["time"].values)

    # can't run again using a different time shift (the default)
    with pytest.raises(ValueError, match="has already been scaled"):
        Cocip(met, rad)


def test_cocip_time_handling(fl: Flight, met: MetDataset, rad: MetDataset) -> None:
    """Check a few aspects of Cocip time handling."""
    params = {
        "max_age": np.timedelta64(4, "h"),
        "process_emissions": False,
        "humidity_scaling": ExponentialBoostHumidityScaling(),
    }
    cocip = Cocip(met, rad=rad.copy(), params=params)

    # Here, the timesteps attribute is not created until eval is called
    cocip.eval(source=fl)
    assert cocip.timesteps.size == 13

    # In general, we won't get equality here. But we do for this particular flight.
    assert cocip.timesteps[-1] == fl["time"][-1] + params["max_age"]
    assert fl["time"][0] < cocip.timesteps[0] < fl["time"][0] + cocip.params["dt_integration"]

    params["dt_integration"] = np.timedelta64(15, "m")
    cocip = Cocip(met, rad=rad.copy(), params=params)
    cocip.eval(source=fl)
    assert len(cocip.timesteps) == 26


def test_eval_setup(fl: Flight, met: MetDataset, rad: MetDataset) -> None:
    """Check `Cocip` logic appearing at start of `eval` method."""
    # required flight parameters
    del fl.attrs["flight_id"]
    cocip = Cocip(
        met=met,
        rad=rad,
        process_emissions=False,
        humidity_scaling=ExponentialBoostHumidityScaling(),
    )
    match = "Source flight does not contain `flight_id` data or attr. Adding `flight_id` of 0"
    with pytest.warns(UserWarning, match=match):
        cocip.eval(source=fl)


def test_flight_overrides(fl: Flight, met: MetDataset, rad: MetDataset) -> None:
    """Ensure flight keys not overwritten by `Cocip`."""
    fl.update(longitude=np.linspace(-29, -32, 20))
    fl.update(latitude=np.linspace(51, 58, 20))
    fl.update(altitude=np.full(20, 10000))

    # Add flight / met properties that override default inputs
    fl2 = fl.copy()
    fl2["true_airspeed"] = np.full(fl2.size, 40)
    fl2["segment_length"] = np.full(fl2.size, 10000)
    fl2["air_temperature"] = np.full(fl2.size, 220)

    # Here, fl1 has gaps > 40000 (default) between segments
    cocip = Cocip(
        met=met.copy(),
        rad=rad.copy(),
        process_emissions=False,
        max_seg_length_m=50000,
        humidity_scaling=ExponentialBoostHumidityScaling(),
    )

    # Eventually the advected waypoints will blow out of bounds
    # Specifically, there is a contrail at 4:30 with latitude larger than 59
    # We acknowledge this here
    with pytest.warns(UserWarning, match="outside of the met interpolation grid"):
        out1 = cocip.eval(fl)
        out2 = cocip.eval(fl2)

    # Because Cocip copies fl, the original isn't modified
    assert "segment_length" not in fl
    assert "true_airspeed" not in fl

    # We see the segment properties on the output
    assert np.all(out1["segment_length"][:-1] > 42000)
    assert np.all(out1["segment_length"][:-1] < 43000)
    assert np.all(out1["true_airspeed"][:-1] > 62)
    assert np.all(out1["true_airspeed"][:-1] < 69)

    # And when segment_length and true_airspeed are already on the flight,
    # the values are unchanged
    assert np.all(out2["segment_length"] == 10000)
    assert np.all(out2["true_airspeed"] == 40)
    assert np.all(out2["air_temperature"] == 220)

    # AND, these variables do impact the model output!
    assert out1["cocip"][18] == 0
    assert out2["cocip"][18] == 1

    # The rest are equal
    filt = np.ones(len(out1), dtype=bool)
    filt[18] = False
    np.testing.assert_array_equal(out1["cocip"][filt], out2["cocip"][filt])

    # Test "waypoint" data provided with flight
    fl3 = fl.copy()
    fl3["waypoint"] = np.arange(len(fl))
    # Using a short max age to avoid advecting out of bounds
    cocip.params["max_age"] = np.timedelta64(1, "h")
    out = cocip.eval(fl3)

    # No nans at positive predictions
    assert out.dataframe[out.dataframe.ef > 0].notna().all().all()

    fl3["waypoint"][14:20] += 1
    with pytest.raises(ValueError, match="non-sequential waypoints"):
        cocip.eval(fl3)


def test_flight_overrides_emissions(
    fl: Flight, met: MetDataset, rad: MetDataset, bada_model: AircraftPerformance
) -> None:
    """Ensure flight keys not overwritten by `Cocip` with emissions modeling."""

    fl4 = fl.copy()
    del fl4.data["aircraft_mass"]  # this will throw a warning, which causes an error
    del fl4.attrs["nvpm_ei_n"]
    nvpm_ei_n = np.linspace(1e15, 2e15, fl.size)
    fl4["nvpm_ei_n"] = nvpm_ei_n

    params = {
        "process_emissions": True,
        "aircraft_performance": bada_model,
        "humidity_scaling": ExponentialBoostHumidityScaling(),
    }
    cocip = Cocip(met=met.copy(), rad=rad.copy(), params=params)
    out4 = cocip.eval(fl4)

    assert np.all(out4["aircraft_mass"] != fl.data["aircraft_mass"])
    assert np.all(out4["nvpm_ei_n"] == nvpm_ei_n)
    assert np.all(out4["engine_efficiency"] == fl4.data["engine_efficiency"])
    assert np.all(out4["fuel_flow"] == fl4.data["fuel_flow"])
    assert np.all(out4["thrust"] == fl4.attrs["thrust"])
    assert out4.attrs["wingspan"] == fl4.attrs["wingspan"]


@pytest.mark.filterwarnings("ignore:distutils Version classes are deprecated")
def test_flight_output(fl: Flight, met: MetDataset, rad: MetDataset) -> None:
    """Check `Cocip` outputs against data in static directory."""
    cocip = Cocip(
        met,
        rad=rad,
        process_emissions=False,
        humidity_scaling=ExponentialBoostHumidityScaling(),
    )
    out = cocip.eval(source=fl)

    assert isinstance(out, Flight)
    assert out is not fl
    assert np.all(out["latitude"] == fl["latitude"])
    assert np.all(out["longitude"] == fl["longitude"])
    assert np.all(out["time"] == fl["time"])
    assert np.all(out.level == out.level)

    assert "rf_net_mean" in cocip.source
    assert "ef" in cocip.source
    assert "cocip" in cocip.source

    assert out.attrs["flight_id"] == fl.attrs["flight_id"]

    # make sure we can output flight dataframe
    # this only works with fastparquet because `age` which is not handled by pyarrow correctly
    out.dataframe.to_parquet(".test.pq", engine="fastparquet")
    assert pathlib.Path(".test.pq").exists()

    # note that fastparquet truncastes to [us],
    # so when we read this back in the last 3 digits are off
    # pyarrow doesn't support timedelta64 past "us", but pandas won't
    # allow casting to "us" in a dataframe - bit of a mess
    df = pd.read_parquet(".test.pq", engine="fastparquet")
    np.testing.assert_allclose(df["contrail_age"].values, out.dataframe["contrail_age"].values)

    # clean up
    pathlib.Path(".test.pq").unlink()

    # make sure we can output flight dataframe using pyarrow with age in seconds
    out.update(contrail_age=out["contrail_age"] / np.timedelta64(1, "s"))
    out.update(time=out["time"].astype("datetime64[us]"))
    out.dataframe.to_parquet(".test2.pq", engine="pyarrow")
    assert pathlib.Path(".test2.pq").exists()

    df = pd.read_parquet(".test2.pq", engine="pyarrow")
    np.testing.assert_allclose(df["contrail_age"].values, out.dataframe["contrail_age"].values)

    # clean up
    pathlib.Path(".test2.pq").unlink()


def test_eval_no_ef(cocip_no_ef: Cocip) -> None:
    """Confirm pinned values of `cocip_no_eval` fixture."""
    # Confirm no nan in output besides the terminal waypoint that is always nan
    assert cocip_no_ef._sac_flight.dataframe.iloc[:-1].notna().all().all()
    assert cocip_no_ef._downwash_flight.dataframe.iloc[:-1].notna().all().all()

    # Pin sizes
    assert cocip_no_ef._sac_flight.size == 20
    assert cocip_no_ef._downwash_flight.size == 19

    assert "rf_net_mean" in cocip_no_ef.source
    assert "ef" in cocip_no_ef.source
    assert "persistent_1" in cocip_no_ef.source

    # Flight doesn't produce any persistent contrails
    assert np.all(cocip_no_ef.source["ef"] == 0)
    assert np.all(cocip_no_ef.source["cocip"] == 0)
    assert np.all(cocip_no_ef.source["contrail_age"] == np.timedelta64(0, "ns"))

    # BUT it does satisfy the SAC
    assert np.all(cocip_no_ef.source["sac"] == 1)

    # And initial persistent
    assert np.all(cocip_no_ef.source["persistent_1"][:-1] == 1)
    assert cocip_no_ef.source["persistent_1"][-1] == 0


def test_eval_persistent(cocip_persistent: Cocip) -> None:
    """Confirm pinned values of `cocip_persistent` fixture."""
    assert cocip_persistent.timesteps.size == 11

    assert "rf_net_mean" in cocip_persistent.source
    assert "ef" in cocip_persistent.source
    assert "cocip" in cocip_persistent.source
    assert np.nansum(cocip_persistent.source["cocip"]) > 0
    assert cocip_persistent.contrail is not None

    # # output json when algorithm has been adjusted
    # cocip_persistent.source.dataframe.to_json(
    #     get_static_path("cocip-flight-output.json"),
    #     indent=2,
    #     orient="records",
    #     date_unit="ns",
    #     double_precision=15,
    # )
    # cocip_persistent.contrail.to_json(
    #     get_static_path("cocip-contrail-output.json"),
    #     indent=2,
    #     orient="records",
    #     date_unit="ns",
    #     double_precision=15,
    # )

    flight_output = pd.read_json(get_static_path("cocip-flight-output.json"), orient="records")
    contrail_output = pd.read_json(get_static_path("cocip-contrail-output.json"), orient="records")

    flight_output["time"] = pd.to_datetime(flight_output["time"])
    flight_output["contrail_age"] = pd.to_timedelta(flight_output["contrail_age"])
    contrail_output["time"] = pd.to_datetime(contrail_output["time"])
    contrail_output["formation_time"] = pd.to_datetime(contrail_output["formation_time"])
    contrail_output["age"] = pd.to_timedelta(contrail_output["age"])
    contrail_output["dt_integration"] = pd.to_timedelta(contrail_output["dt_integration"])

    rtol = 1e-4
    for key in flight_output:
        if key in ["time", "flight_id"]:
            np.testing.assert_array_equal(cocip_persistent.source[key], flight_output[key])
            continue
        np.testing.assert_allclose(
            cocip_persistent.source.get_data_or_attr(key),
            flight_output[key],
            err_msg=key,
            rtol=rtol,
        )

    pd.testing.assert_frame_equal(
        cocip_persistent.contrail.reset_index(drop=True),
        contrail_output.reset_index(drop=True),
        check_dtype=False,
        check_like=True,  # ignore column order
        check_exact=False,
        rtol=rtol,
    )


def test_eval_persistent2(cocip_persistent2: Cocip) -> None:
    """Confirm pinned values of ``cocip_persistent2`` fixture."""
    assert cocip_persistent2.timesteps.size == 16

    assert "rf_net_mean" in cocip_persistent2.source
    assert "ef" in cocip_persistent2.source
    assert "cocip" in cocip_persistent2.source
    assert np.all(np.isfinite(cocip_persistent2.source["ef"]))
    assert np.all(np.isfinite(cocip_persistent2.source["cocip"]))

    assert np.sum(cocip_persistent2.source["cocip"]) > 0
    assert cocip_persistent2.contrail is not None

    # # output json when algorithm has been adjusted
    # cocip_persistent2.source.dataframe.to_json(
    #     get_static_path("cocip-flight-output2.json"),
    #     indent=2,
    #     orient="records",
    #     date_unit="ns",
    #     double_precision=15,
    # )
    # cocip_persistent2.contrail.to_json(
    #     get_static_path("cocip-contrail-output2.json"),
    #     indent=2,
    #     orient="records",
    #     date_unit="ns",
    #     double_precision=15,
    # )

    # confirm all discontinuous waypoints have an "ef" of 0 and an age of 0
    continuous = cocip_persistent2.contrail["continuous"].to_numpy()
    assert np.all(cocip_persistent2.contrail["ef"][~continuous] == 0)
    assert np.all(cocip_persistent2.contrail["age"][~continuous] == np.timedelta64(0, "ns"))

    flight_output = pd.read_json(get_static_path("cocip-flight-output2.json"), orient="records")
    contrail_output = pd.read_json(get_static_path("cocip-contrail-output2.json"), orient="records")

    flight_output["time"] = pd.to_datetime(flight_output["time"])
    flight_output["contrail_age"] = pd.to_timedelta(flight_output["contrail_age"])
    contrail_output["time"] = pd.to_datetime(contrail_output["time"])
    contrail_output["formation_time"] = pd.to_datetime(contrail_output["formation_time"])
    contrail_output["age"] = pd.to_timedelta(contrail_output["age"])
    contrail_output["dt_integration"] = pd.to_timedelta(contrail_output["dt_integration"])

    rtol = 1e-3
    for key in flight_output:
        if key in ["time", "flight_id"]:
            assert np.all(cocip_persistent2.source[key] == flight_output[key])
            continue
        if key == "level":
            np.testing.assert_allclose(
                cocip_persistent2.source.level, flight_output[key], err_msg=key
            )
            continue

        np.testing.assert_allclose(
            cocip_persistent2.source[key], flight_output[key], err_msg=key, rtol=rtol
        )

    pd.testing.assert_frame_equal(
        cocip_persistent2.contrail.reset_index(drop=True),
        contrail_output.reset_index(drop=True),
        check_dtype=False,
        check_like=True,  # ignore column order
        check_exact=False,
        rtol=rtol,
    )


def test_xarray_contrail(cocip_persistent: Cocip) -> None:
    """Confirm expected contrail output on attribute `contrail_dataset`."""
    assert cocip_persistent.contrail_dataset["longitude"].shape == (8, 12)

    # The contrail_dataset['timestep'] does not necessarily coincide with cocip.timesteps
    assert cocip_persistent.contrail_dataset["timestep"].size == 8
    assert cocip_persistent.contrail_dataset["waypoint"].size == 12


def test_emissions(met: MetDataset, rad: MetDataset, bada_model: AircraftPerformance) -> None:
    """Test `Cocip` use of `Emissions` model."""
    # demo synthetic flight for BADA
    flight_parameters = {"flight_id": "test BADA/EDB"}
    aircraft = Aircraft(aircraft_type="A320")

    # Example flight
    df = pd.DataFrame()
    df["longitude"] = np.linspace(-21, -40, 50)
    df["latitude"] = np.linspace(51, 59, 50)
    df["altitude"] = np.linspace(10900, 11000, 50)
    df["time"] = pd.date_range("2019-01-01T00:15:00", "2019-01-01T02:30:00", periods=50)

    fl = Flight(data=df, aircraft=aircraft, attrs=flight_parameters)
    # Confirm all initial segment lengths are below the cocip
    # max_seg_length_m default param
    assert Cocip.default_params.max_seg_length_m == 40000
    assert np.all(fl.segment_length()[:-1] < 40000)

    params = {
        "process_emissions": True,
        "max_age": np.timedelta64(3, "h"),
        "aircraft_performance": bada_model,
        "humidity_scaling": ExponentialBoostHumidityScaling(),
    }
    cocip = Cocip(met, rad=rad, params=params)
    cocip.eval(source=fl)

    assert "engine_efficiency" in cocip.source
    assert np.all(cocip.source["engine_efficiency"][:-1] < 0.37)
    assert np.all(cocip.source["engine_efficiency"][:-1] > 0.28)

    assert "fuel_flow" in cocip.source
    assert np.all(cocip.source["fuel_flow"][:-1] < 0.73)
    assert np.all(cocip.source["fuel_flow"][:-1] > 0.72)

    assert "nvpm_ei_n" in cocip.source
    assert np.all(cocip.source["nvpm_ei_n"][:-1] < 1.20e15)
    assert np.all(cocip.source["nvpm_ei_n"][:-1] > 1.05e15)

    assert "thrust" in cocip.source
    assert np.all(cocip.source["thrust"][:-1] < 81000)
    assert np.all(cocip.source["thrust"][:-1] > 51000)

    assert "thrust_setting" in cocip.source
    assert np.all(cocip.source["thrust_setting"][:-1] < 0.51)
    assert np.all(cocip.source["thrust_setting"][:-1] > 0.47)


def test_intermediate_columns_in_flight_output(cocip_persistent: Cocip) -> None:
    """Ensure returned `Flight` instance has expected columns."""
    for col in ["rhi", "width", "T_critical_sac", "rh_critical_sac", "sac"]:
        assert col in cocip_persistent.source
        assert np.all(np.isfinite(cocip_persistent.source[col]))


# ------
# Output
# ------


def test_flight_statistics(cocip_persistent: Cocip) -> None:
    """Test `Cocip.output_flight_statistics`."""
    stats = cocip_persistent.output_flight_statistics()
    assert isinstance(stats, pd.Series)

    # update output stats when algorithm is adjusted
    # stats.to_json(get_static_path("cocip-flight-statistics.json"), indent=2)

    # compare each value after being processed through json
    output_stats = json.loads(stats.to_json())

    _path = get_static_path("cocip-flight-statistics.json")
    with open(_path, "r") as f:
        opened_stats = json.load(f)

    for k in output_stats:
        if opened_stats[k] is None:
            continue
        if isinstance(opened_stats[k], str):
            assert output_stats[k] == opened_stats[k]
            continue
        np.testing.assert_allclose(output_stats[k], opened_stats[k], err_msg=k, rtol=1e-5)


def test_grid_cirrus(cocip_persistent: Cocip) -> None:
    """Test `grid_cirrus.cirrus_summary_statistics` function."""
    df_contrails = cocip_persistent.contrail
    assert isinstance(df_contrails, pd.DataFrame)
    df_contrails["flight_id"] = cocip_persistent.source.attrs["flight_id"]

    grid_summary = grid_cirrus.cirrus_summary_statistics(
        df_contrails,
        cocip_persistent.met.data["fraction_of_cloud_cover"].isel(time=slice(0, 8)),
        cocip_persistent.met.data["tau_cirrus"].isel(time=slice(0, 8)),
    )

    # summary json output for comparison
    # grid_summary.to_json(
    #     get_static_path("cocip-output-grid-cirrus-summary.json"), indent=2, orient="records"
    # )

    # load summary json output for comparison
    grid_summary_previous = pd.read_json(get_static_path("cocip-output-grid-cirrus-summary.json"))

    for key in grid_summary:
        assert np.allclose(grid_summary[key], grid_summary_previous[key])


def test_contrail_edges(cocip_persistent: Cocip) -> None:
    """Test contrail edges methods."""
    df_contrail = cocip_persistent.contrail
    assert df_contrail is not None

    (
        df_contrail["lon_edge_l"],
        df_contrail["lat_edge_l"],
        df_contrail["lon_edge_r"],
        df_contrail["lat_edge_r"],
    ) = contrail_properties.contrail_edges(
        df_contrail["longitude"].to_numpy(),
        df_contrail["latitude"].to_numpy(),
        df_contrail["sin_a"].to_numpy(),
        df_contrail["cos_a"].to_numpy(),
        df_contrail["width"].to_numpy(),
    )

    # summary json output for comparison
    # df_contrail[["lon_edge_l", "lat_edge_l", "lon_edge_r", "lat_edge_r"]].to_json(
    #     get_static_path("cocip-output-contrail-edges.json"), indent=2, orient="records"
    # )

    # load summary json output for comparison
    df_contrail_previous = pd.read_json(get_static_path("cocip-output-contrail-edges.json"))

    keys = ["lon_edge_l", "lat_edge_l", "lon_edge_r", "lat_edge_r"]
    np.testing.assert_allclose(df_contrail[keys].to_numpy(), df_contrail_previous[keys].to_numpy())


def test_cocip_fleet(fl: Flight, met: MetDataset, rad: MetDataset):
    """Confirm that Cocip runs in "fleet" mode.

    Simple smoke test achieved by concatenating several copies of fl together.
    """
    initial_keys = set(fl)
    fls = [fl.copy() for _ in range(10)]
    cocip = Cocip(
        met=met,
        rad=rad,
        process_emissions=False,
        humidity_scaling=ExponentialBoostHumidityScaling(),
    )
    with pytest.raises(ValueError, match="Duplicate `flight_id`"):
        cocip.eval(fls)

    # confirm nothing has been mutated (using default copy=True)
    for fl in fls:
        assert set(fl) == initial_keys

    for i, fl in enumerate(fls):
        fl.attrs.update(flight_id=f"flight_id_{i}")

    out = cocip.eval(source=fls)
    assert isinstance(out, list)
    for fl in out:
        assert isinstance(fl, Flight)

        # Removing anything that is different between flights
        del fl.data["flight_id"]
        pd.testing.assert_frame_equal(fl.dataframe, out[0].dataframe)


# This is not explicitly necessary, but I've added for benchmarking
@pytest.mark.skipif(True, reason="Profiling test off")
def test_cocip_performance(met: MetDataset, rad: MetDataset) -> None:
    """Test the performance of the CoCip algorithm over a wide domain."""
    # Demo flight
    n = 10000
    flight_parameters = {
        "flight_id": "test",
        "aircraft_type": "A380",
        "wingspan": 48,
        "bc_ei_n": 1.897462e15,
    }
    df = pd.DataFrame()
    df["longitude"] = np.linspace(-39, -21, n)
    df["latitude"] = np.linspace(50, 59, n)
    df["altitude"] = np.linspace(10900, 10900, n)
    df["time"] = pd.date_range("2019-01-01T00:15:00", "2019-01-01T02:05:00", periods=n)
    df["engine_efficiency"] = np.linspace(0.32, 0.45, n)  # ope
    df["fuel_flow"] = np.linspace(2.1, 2.4, n)  # kg/s
    df["aircraft_mass"] = np.linspace(154445, 154345, n)  # kg
    df["bc_ei_n"] = np.linspace(1.8e15, 1.9e15, n)  # kg
    fl = Flight(df, **flight_parameters)

    # run CoCip
    params = {
        "max_age": np.timedelta64(5, "h"),
        "process_emissions": False,
        "verbose_outputs": True,
    }

    # time 10 cocip calls
    start = pythontime.perf_counter()

    for i in range(10):
        cocip = Cocip(met.copy(), rad=rad.copy(), params=params)
        _ = cocip.eval(source=fl)

    duration = pythontime.perf_counter() - start

    # most recently 5.0 seconds on macbook air M1 2020 with 16GB ram
    print(f"10 Cocip runs took {duration:.2f} seconds")


def test_cocip_gfs(fl: Flight, met_gfs: MetDataset, rad_gfs: MetDataset) -> None:
    """Confirm pinned values of `cocip_persistent` fixture."""

    # met data is on 2022-01-01 - need to scale time to here
    dt = np.datetime64("2022-01-01 00:15:00") - fl["time"][0]
    fl_gfs = fl.copy()
    fl_gfs.update(time=fl_gfs["time"] + dt)
    fl_gfs.update(longitude=np.linspace(-29, -32, 20))
    fl_gfs.update(latitude=np.linspace(56, 58, 20))
    fl_gfs.update(altitude=np.linspace(10900, 10900, 20))

    # run CoCip
    # If process_emissions=True, BADA will run and overwrite values for
    # engine_efficiency, ...
    params = {
        "max_age": np.timedelta64(3, "h"),
        "process_emissions": False,
        "verbose_outputs": True,
    }
    cocip = Cocip(met=met_gfs, rad=rad_gfs, params=params)
    out = cocip.eval(source=fl_gfs)

    assert isinstance(out, Flight)

    assert "sac" in out
    assert "rhi" in out
    assert "ef" in out
    assert np.nansum(out["sac"]) > 0

    # TODO: find a test case where EF > 0 with GFS. Unfortunately seems
    # like GFS really doesn't do ISSR regions
    # assert np.nansum(out["ef"]) > 0


def test_cocip_filtering(fl: Flight, met: MetDataset, rad: MetDataset):
    """Confirm Cocip runs with non-default filtering flags.

    Check model parameters:
    - filter_sac
    - filter_initially_persistent
    - persistent_buffer
    """
    # Boost air temperature for more interesting SAC dynamics
    met.data["air_temperature"] += 10

    scaler = hs.ExponentialBoostHumidityScaling(rhi_adj=0.8)
    params = {"met": met, "rad": rad, "process_emissions": False, "humidity_scaling": scaler}
    cocip = Cocip(**params)
    cocip.eval(fl)
    assert cocip._sac_flight.size == 18
    assert cocip._downwash_flight.size == 10
    assert len(cocip.contrail) == 0

    cocip = Cocip(**params, filter_sac=False)
    with pytest.warns(UserWarning, match="Manually overriding SAC filter"):
        cocip.eval(fl)
    assert cocip._sac_flight.size == 20
    assert cocip._downwash_flight.size == 10
    assert len(cocip.contrail) == 0

    cocip = Cocip(**params, filter_initially_persistent=False)
    with pytest.warns(UserWarning, match="Manually overriding initially persistent filter"):
        cocip.eval(fl)
    assert cocip._sac_flight.size == 18
    assert cocip._downwash_flight.size == 18
    assert len(cocip.contrail) == 0

    cocip = Cocip(
        **params,
        filter_initially_persistent=False,
        persistent_buffer=np.timedelta64(1, "h"),
    )
    with pytest.warns(UserWarning, match="Manually overriding initially persistent filter"):
        cocip.eval(fl)
    assert len(cocip.contrail) == 18 * 2
    assert "end_of_life" in cocip.contrail

    cocip = Cocip(
        **params,
        filter_initially_persistent=False,
        persistent_buffer=np.timedelta64(2, "h"),
    )
    with pytest.warns(UserWarning, match="Manually overriding initially persistent filter"):
        cocip.eval(fl)
    assert len(cocip.contrail) == 18 * 4
    assert "end_of_life" in cocip.contrail


def test_cocip_no_persistence_ef_fill_value(fl: Flight, met: MetDataset, rad: MetDataset):
    """Confirm that EF is filled with 0 for in-domain and nan for out-of-domain waypoints."""
    # Cherrypick some trajectory that is half in-domain and half out-of-domain
    # And make sure SAC is not satisfied for in-domain waypoints
    fl.update(altitude=np.linspace(11000, 13000, 20))
    fl.update(longitude=np.linspace(-39, -37, 20))
    fl.update(latitude=np.full(20, 51))

    # The first 8 waypoints are in-domain, the last 12 are out-of-domain
    np.testing.assert_array_equal(fl.coords_intersect_met(met), [True] * 8 + [False] * 12)

    params = {
        "met": met,
        "rad": rad,
        "process_emissions": False,
        "humidity_scaling": ExponentialBoostHumidityScaling(),
    }
    cocip = Cocip(**params)
    out = cocip.eval(fl)

    # Check that ef and contrail_age are 0 for in-domain and nan for out-of-domain waypoints
    np.testing.assert_array_equal(out["sac"], [0] * 8 + [np.nan] * 12)
    np.testing.assert_array_equal(out["ef"], [0] * 8 + [np.nan] * 12)
    np.testing.assert_array_equal(
        out["contrail_age"], [np.timedelta64(0, "s")] * 8 + [np.timedelta64("nat")] * 12
    )

    # This key was removed
    assert "_met_intersection" not in out


def test_radiative_heating_effects_param(fl: Flight, met: MetDataset, rad: MetDataset):
    """Run Cocip with the radiative_heating_effects parameter.

    The purpose of this test is to show that real differences accrue when using the
    radiative_heating_effects parameter.
    """

    fl.update(longitude=np.linspace(-29, -32, 20))
    fl.update(latitude=np.linspace(56, 57, 20))
    fl.update(altitude=np.linspace(10900, 10900, 20))

    params = {
        "max_age": np.timedelta64(90, "m"),  # keep short to avoid blowing out of bounds
        "dt_integration": np.timedelta64(10, "m"),
        "process_emissions": False,
        "humidity_scaling": ExponentialBoostLatitudeCorrectionHumidityScaling(),
    }

    # Artifically shift time to get some SDR
    met.data = met.data.assign_coords(time=met.data["time"] + np.timedelta64(12, "h"))
    rad.data = rad.data.assign_coords(time=rad.data["time"] + np.timedelta64(12, "h"))
    fl.update(time=fl["time"] + np.timedelta64(12, "h"))

    cocip = Cocip(met, rad=rad, params=params)
    fl1 = cocip.eval(source=fl)
    contrail1 = cocip.contrail
    assert not cocip.params["radiative_heating_effects"]
    fl2 = cocip.eval(source=fl, radiative_heating_effects=True)
    assert cocip.params["radiative_heating_effects"]
    contrail2 = cocip.contrail

    # Compare output between two model runs
    expected = {"d_heat_rate", "cumul_differential_heat", "heat_rate", "cumul_heat"}
    assert set(contrail2.columns).difference(contrail1.columns) == expected

    # Pretty massive difference in EF
    assert fl1["ef"].sum() == pytest.approx(7.3e12, rel=0.1)
    assert fl2["ef"].sum() == pytest.approx(10.1e12, rel=0.1)

    # Not nonzero in the same places!
    filt1 = fl1["ef"] != 0
    filt2 = fl2["ef"] != 0
    assert np.all(filt1 >= filt2)
    assert filt1.sum() == 10
    assert filt2.sum() == 7


def test_radiative_heating_effects():
    """Test radiative heating implementation against precomputed values from Schumann.

    We're somewhat loose with typing in this test. In CoCiP, all of the variables
    defined below are numpy ndarrays. Here, for convenience, we take some as floats
    where it doesn't introduce TypeErrors.
    """
    air_temperature = 217.542
    rhi = 1.1024
    rho_air = 0.4730
    r_ice_vol = 0.000016687

    width = 22731
    depth = np.array([3381.4])
    sigma_yz = 7217929.7
    area_eff = contrail_properties.plume_effective_cross_sectional_area(width, depth, sigma_yz)
    depth_eff = contrail_properties.plume_effective_depth(width, area_eff)
    dT_dz = np.array([0.0023096])
    age_s = 46260

    tau_cirrus = 0.1486182
    sd0 = 1362.215
    sdr = 339.031
    rsr = np.array([145.401])
    olr = 217.848

    # Differential heating rate will always be zero if tau_contrail is zero
    tau_contrail = 0.0
    d_heat_rate = radiative_heating.differential_heating_rate(
        air_temperature,
        rhi,
        rho_air,
        r_ice_vol,
        depth_eff,
        tau_contrail,
        tau_cirrus,
        sd0,
        sdr,
        rsr,
        olr,
    )
    assert d_heat_rate.item() == 0.0

    # Test case -- heating rate - agreement with Fortran values
    tau_contrail = 0.150
    heat_rate = radiative_heating.heating_rate(
        air_temperature,
        rhi,
        rho_air,
        r_ice_vol,
        depth_eff,
        tau_contrail,
        tau_cirrus,
        sd0,
        sdr,
        rsr,
        olr,
    )
    np.testing.assert_allclose(heat_rate, 1.7841781452750816e-06)

    # Test case -- differential heating rate - agreement with Fortran values
    tau_contrail = 0.150
    d_heat_rate = radiative_heating.differential_heating_rate(
        air_temperature,
        rhi,
        rho_air,
        r_ice_vol,
        depth_eff,
        tau_contrail,
        tau_cirrus,
        sd0,
        sdr,
        rsr,
        olr,
    )
    np.testing.assert_allclose(d_heat_rate, -3.853027307117556e-06)

    cumul_rad_heat = -d_heat_rate * age_s  # Cumulative heat should be a positive value
    eff_heat_rate = radiative_heating.effective_heating_rate(
        d_heat_rate, cumul_rad_heat, dT_dz, depth
    )
    np.testing.assert_allclose(eff_heat_rate, -1.6819834787264378e-07)

    ratio = eff_heat_rate / d_heat_rate
    assert 0 < ratio < 1


@pytest.mark.parametrize("max_altitude", [11200, None])
def test_max_altitude_param(fl: Flight, met: MetDataset, rad: MetDataset, max_altitude):
    """Confirm that the max_altitude parameter can be disabled."""
    params = {
        "max_age": np.timedelta64(10, "m"),
        "dt_integration": np.timedelta64(5, "m"),
        "process_emissions": False,
        "humidity_scaling": ExponentialBoostHumidityScaling(rhi_adj=0.3),
        "max_altitude_m": max_altitude,
    }

    fl.update(longitude=np.linspace(-29, -32, 20))
    assert fl.altitude.min() == 11000
    assert fl.altitude.max() == 11500

    cocip = Cocip(met, rad=rad, params=params)
    ef = cocip.eval(source=fl)["ef"]

    # They're not all persistent because of continuity conventions
    if max_altitude is None:
        assert np.count_nonzero(ef) == 11
    else:
        assert np.count_nonzero(ef) == 6



"""Test ACCF model"""


from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

from pycontrails import Flight, MetDataset
from pycontrails.models.accf import ACCF


def _is_climaccf_available():
    try:
        import climaccf  # noqa: F401

        return True
    except ModuleNotFoundError:
        return False


@pytest.mark.skipif(not _is_climaccf_available(), reason="climaccf package not available")
def test_accf_default(met_accf_pl: MetDataset, met_accf_sl: MetDataset) -> None:
    """Test Default accf algorithm."""

    # little hack to avoid warning
    del met_accf_pl.attrs["history"]
    met_accf_pl.attrs["met_source"] = "not_ecmwf"
    del met_accf_sl.attrs["history"]
    met_accf_sl.attrs["met_source"] = "not_ecmwf"

    accf = ACCF(met_accf_pl, met_accf_sl)

    n = 10000
    longitude = np.linspace(45, 75, n) + np.linspace(0, 1, n)
    latitude = np.linspace(50, 60, n) - np.linspace(0, 1, n)
    level = np.full_like(longitude, 225)

    start = np.datetime64("2022-11-11")
    time = pd.date_range(start, start + np.timedelta64(90, "m"), periods=n)
    fl = Flight(
        longitude=longitude,
        latitude=latitude,
        level=level,
        time=time,
        aircraft_type="B737",
        flight_id=17,
    )

    out = accf.eval(fl)

    assert np.all(np.isfinite(out["aCCF_NOx"]))
